import inspect
import os
from pathlib import Path
from typing import Tuple, Union, List, Dict

import numpy as np
import pymem3dg as dg
from h5py import Dataset
import matplotlib.pyplot as plt
import matplotlib.animation as animation
from process_bigraph import Composite


def reflect_forces(particles, vertex_matrix, force_vectors, time_step):
    for particle in particles:
        # Find nearest vertex
        distances = np.linalg.norm(vertex_matrix - particle.position, axis=1)
        nearest_vertex_idx = np.argmin(distances)

        # Reflect force onto the particle
        force = -force_vectors[nearest_vertex_idx]  # Reaction force
        particle.velocity += force * time_step / particle.mass


def calculate_preferred_area(v_preferred: float, convert_to_micro: bool = False) -> float:
    V_preferred = v_preferred
    V_preferred = V_preferred * 1e15 if convert_to_micro else V_preferred  # Convert liters to μm³
    r_preferred = (3 * V_preferred / (4 * np.pi)) ** (1/3)
    A_preferred = 4 * np.pi * r_preferred**2

    return A_preferred


def new_parameters(param_spec: Dict):
    parameters = dg.Parameters()
    for attribute_name, attribute_spec in param_spec.items():  # ie: adsorption, aggregation, bending, etc
        attribute = getattr(parameters, attribute_name)
        if isinstance(attribute_spec, dict):
            for name, value in attribute_spec.items():
                setattr(attribute, name, value)
        else:
            setattr(parameters, attribute_name, attribute_spec)
    return parameters


def parse_parameters(parameters: dg.Parameters):
    # parse parameters for iteration
    param_data = {}
    for param_attr_name in dir(parameters):
        if not param_attr_name.startswith("__"):
            param_data[param_attr_name] = {}
            attr = getattr(parameters, param_attr_name)
            attr_props = {}
            for inner_name in dir(attr):
                if not inner_name.startswith("__"):
                    inner_attr = getattr(attr, inner_name)
                    if not callable(inner_attr) or not inspect.isbuiltin(inner_attr):
                        attr_props[inner_name] = inner_attr

            param_data[param_attr_name] = attr_props

    return param_data


def parse_ply(file_path: os.PathLike[str]) -> Tuple[np.ndarray, np.ndarray]:
    with open(file_path, 'r') as file:
        lines = file.readlines()
    vertex_matrix = []
    face_matrix = []
    vertex_count = 0
    face_count = 0
    in_vertex_section = False
    in_face_section = False
    for line in lines:
        line = line.strip()
        if line.startswith("element vertex"):
            vertex_count = int(line.split()[-1])
        elif line.startswith("element face"):
            face_count = int(line.split()[-1])
        elif line == "end_header":
            in_vertex_section = True
            continue

        # vertices
        if in_vertex_section:
            if len(vertex_matrix) < vertex_count:
                vertex_matrix.append(list(map(float, line.split())))
                if len(vertex_matrix) == vertex_count:
                    in_vertex_section = False
                    in_face_section = True
            continue

        # faces
        if in_face_section:
            if len(face_matrix) < face_count:
                parts = list(map(int, line.split()))
                face_matrix.append(parts[1:])  # Skip the first number (vertex count)
                if len(face_matrix) == face_count:
                    break

    return np.array(face_matrix, dtype=np.uint64), np.array(vertex_matrix, dtype=np.float64)


def extract_data(
        dataset: Dataset,
        data_name: str,
        return_last: bool = True,
        use_numpy: bool = False
) -> List[Union[List[float], List[List[float]]]]:
    traj_data = dataset.groups['Trajectory'].variables[data_name][:]
    traj_data = traj_data.tolist() if use_numpy else traj_data
    outputs = []
    for data_t in traj_data:
        outputs_t = [
            [data_t[i], data_t[i + 1], data_t[i + 2]] for i in range(0, len(data_t), 3)
        ]
        outputs.append(outputs_t)
    return outputs[-1] if return_last else outputs


def extract_last_data(
        dataset: Dataset,
        data_name: str
) -> List[Union[List[float], List[List[float]]]]:
    traj_data = dataset.groups['Trajectory'].variables[data_name][:]
    traj_data = traj_data.tolist()

    return traj_data[-1]


def generate_faces(width, height):
    """
    Generate face data for a grid mesh.
    Each face is represented as a list of vertex indices.

    Parameters:
        width (int): Number of vertices along the grid width.
        height (int): Number of vertices along the grid height.

    Returns:
        list of list: Each inner list represents a triangular face.
    """

    faces = []
    for i in range(height - 1):
        for j in range(width - 1):
            # Vertex indices of the current grid cell (as a quad)
            v0 = i * width + j
            v1 = v0 + 1
            v2 = v0 + width
            v3 = v2 + 1

            # Split the quad into two triangles
            faces.append([v0, v1, v2])  # Triangle 1
            faces.append([v1, v3, v2])  # Triangle 2
    return faces


def save_mesh_to_ply(vertices, faces, output_path):
    """
    Save vertex and face data to a PLY file.

    Parameters:
        vertices (list of tuples): List of (x, y, z) coordinates for vertices.
        faces (list of lists): List of faces, each represented by a list of vertex indices.
        output_path (str): Path to save the .ply file.
    """
    num_vertices = len(vertices)
    num_faces = len(faces)

    # PLY header
    header = f"""ply
    format ascii 1.0
    element vertex {num_vertices}
    property float x
    property float y
    property float z
    element face {num_faces}
    property list uint8 int32 vertex_indices
    end_header
    """

    # Write the PLY file
    with open(output_path, 'w') as ply_file:
        # Write header
        ply_file.write(header)

        # Write vertex data
        for x, y, z in vertices:
            ply_file.write(f"{x} {y} {z}\n")

        # Write face data
        for face in faces:
            face_str = f"{len(face)} " + " ".join(map(str, face))
            ply_file.write(f"{face_str}\n")


def get_vertex_coordinates(outputDir: Path) -> np.ndarray:
    data = Dataset(str(outputDir / "traj.nc"), 'r')

    # get data parameters
    variables = data.groups['Trajectory'].variables

    # get the coordinates from groups/trajectory.variables
    return data.groups['Trajectory'].variables['coordinates'][:]


def get_axis_vertices(x: np.ndarray, t) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    vertices_x = x[t][::3]
    vertices_y = x[t][1::3]
    vertices_z = x[t][2::3]
    return vertices_x, vertices_y, vertices_z


def format_vertices(x: np.ndarray, t: int, geo: dg.Geometry):
    return np.reshape(x[t], geo.getVertexMatrix().shape)


def get_animation(x: np.ndarray) -> animation.FuncAnimation:
    # Create figure and 3D axis
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')

    n_steps = x.shape[0]
    # Initialize scatter plot with empty data
    scat = ax.scatter([], [], [])

    vertices_x, vertices_y, vertices_z = get_axis_vertices(x, 0)

    # Set axis limits (you may need to adjust based on your data)
    ax.set_xlim(np.min(vertices_x), np.max(vertices_x))
    ax.set_ylim(np.min(vertices_y), np.max(vertices_y))
    ax.set_zlim(np.min(vertices_z), np.max(vertices_z))
    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_zlabel("Z")

    # Update function for animation
    def update(frame):
        ax.set_title(f"Time step: {frame}")

        # Extract x, y, z coordinates for the current frame
        xs = x[frame][::3]
        ys = x[frame][1::3]
        zs = x[frame][2::3]

        # Update scatter plot
        scat._offsets3d = (xs, ys, zs)  # Special way to update 3D scatter plots

        return scat,

    # Create animation
    return animation.FuncAnimation(fig, update, frames=n_steps, interval=50)


def get_times(sim: Composite):
    results = sim.gather_results()[('emitter',)]
    return np.array([t for t in range(len(results))])


def get_vertices(sim: Composite, time_index: int) -> np.ndarray:
    results = sim.gather_results()[('emitter',)]
    n_vertices = len(results[0]['geometry']['vertices'])
    return np.array(results[time_index]['geometry']['vertices'][:n_vertices])





from typing import Dict, Union, Optional

from pandas import DataFrame
from basico import (
    load_model,
    get_species,
    get_parameters,
    get_reactions,
    set_species,
    run_time_course,
    get_compartments,
    new_model,
    set_reaction_parameters,
    add_reaction,
    set_reaction,
    set_parameters,
    add_parameter
)
from process_bigraph import Process

from bsp.data_model.sed import SedModel
from bsp.schemas.config import TimeCourseConfig
from bsp.utils.helpers import fetch_biomodel
from bsp.processes.sed_process import SedUTCProcess


class CopasiProcess(Process):
    """ODE component of the dfba hybrid using COPASI(basico). TODO: Generalize this to use any ode sim."""
    config_schema = {
        'model': SedModel
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)
        self.model = load_model(self.config['model']['model_source'])
        self.reaction_names = get_reactions(model=self.model).index.tolist()
        self.species_names = get_species(model=self.model).index.tolist()

    def initial_state(self):
        initial_concentrations = {
            species_name: get_species(species_name, model=self.model).initial_concentration[0]
            for species_name in self.species_names
        }

        initial_derivatives = {
            rxn_id: get_reactions(rxn_id, model=self.model).flux[0]
            for rxn_id in self.reaction_names
        }

        return {
            'species_concentrations': initial_concentrations,
            'reaction_fluxes': initial_derivatives,
            'time': 0.0
        }

    def inputs(self):
        concentrations_type = {
            name: 'float' for name in self.species_names
        }
        return {
            # 'species_concentrations': concentrations_type,
            'species_counts': {
                name: 'float' for name in self.species_names
            },
            'time': 'float'
        }

    def outputs(self):
        concentrations_type = {
            name: 'float' for name in self.species_names
        }

        reaction_fluxes_type = {
            reaction_name: 'float' for reaction_name in self.reaction_names
        }

        return {
            'species_concentrations': concentrations_type,
            'reaction_fluxes': reaction_fluxes_type,
            'time': 'float'
        }

    def update(self, inputs, interval):
        # spec_data_k = inputs['species_concentrations']
        spec_data_k = inputs['species_counts']
        for cat_id, value in spec_data_k.items():
            # set_type = 'concentration'
            set_type = "count"
            species_config = {
                'name': cat_id,
                'model': self.model,
                set_type: value
            }
            set_species(**species_config)

        # run model for "interval" length; we only want the state at the end
        tc = run_time_course(
            start_time=inputs['time'],
            duration=interval,
            update_model=True,
            model=self.model
        )

        results = {'time': interval}
        results['species_concentrations'] = {
            mol_id: float(get_species(
                name=mol_id,
                exact=True,
                model=self.model
            ).concentration[0])
            for mol_id in self.species_names
        }

        results['reaction_fluxes'] = {
            rxn_id: float(get_reactions(
                name=rxn_id,
                model=self.model
            ).flux[0])
            for rxn_id in self.reaction_names
        }

        return results


class _CopasiProcess(Process):
    """ODE component of the dfba hybrid using COPASI(basico). TODO: Generalize this to use any ode sim."""
    config_schema = {
        'model': SedModel
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)
        self.model = load_model(self.config['model']['model_source'])
        self.reaction_names = get_reactions(model=self.model).index.tolist()
        self.species_names = get_species(model=self.model).index.tolist()

    def initial_state(self):
        initial_concentrations = {
            species_name: get_species(species_name, model=self.model).initial_concentration[0]
            for species_name in self.species_names
        }

        initial_derivatives = {
            rxn_id: get_reactions(rxn_id, model=self.model).flux[0]
            for rxn_id in self.reaction_names
        }

        return {
            'species_concentrations': initial_concentrations,
            'reaction_fluxes': initial_derivatives,
            'time': 0.0
        }

    def inputs(self):
        concentrations_type = {
            name: 'float' for name in self.species_names
        }
        return {
            # 'species_concentrations': concentrations_type,
            'species_counts': {
                name: 'float' for name in self.species_names
            },
            'time': 'float'
        }

    def outputs(self):
        concentrations_type = {
            name: 'float' for name in self.species_names
        }

        reaction_fluxes_type = {
            reaction_name: 'float' for reaction_name in self.reaction_names
        }

        return {
            'species_concentrations': concentrations_type,
            'reaction_fluxes': reaction_fluxes_type,
            'time': 'float'
        }

    def update(self, inputs, interval):
        # spec_data_k = inputs['species_concentrations']
        spec_data_k = inputs['species_counts']
        for cat_id, value in spec_data_k.items():
            # set_type = 'concentration'
            set_type = "count"
            species_config = {
                'name': cat_id,
                'model': self.model,
                set_type: value
            }
            set_species(**species_config)

        # run model for "interval" length; we only want the state at the end
        tc = run_time_course(
            start_time=inputs['time'],
            duration=interval,
            update_model=True,
            model=self.model
        )

        results = {'time': interval}
        results['species_concentrations'] = {
            mol_id: float(get_species(
                name=mol_id,
                exact=True,
                model=self.model
            ).concentration[0])
            for mol_id in self.species_names
        }

        results['reaction_fluxes'] = {
            rxn_id: float(get_reactions(
                name=rxn_id,
                model=self.model
            ).flux[0])
            for rxn_id in self.reaction_names
        }

        return results


# -- fully-spec'd Sed-compliant copasi process --

class SedCopasiProcess(SedUTCProcess):
    config_schema = TimeCourseConfig

    def __init__(self,
                 config: Dict = None,
                 core: Dict = None):
        super().__init__(config, core)

        # insert copasi process model config
        model_source = self.config['model'].get('model_source') or self.config.get('sbml_fp')
        assert model_source is not None, 'You must specify a model source of either a valid biomodel id or model filepath.'
        model_changes = self.config['model'].get('model_changes', {})
        self.model_changes = {} if model_changes is None else model_changes

        # Option A:
        if '/' in model_source:
            self.copasi_model_object = load_model(model_source)
            print('found a filepath')

        # Option B:
        elif 'BIO' in model_source:
            self.copasi_model_object = fetch_biomodel(model_id=model_source)
            print('found a biomodel id')

        # Option C:
        else:
            if not self.model_changes:
                raise AttributeError(
                    """You must pass a source of model changes specifying params, reactions, 
                        species or all three if starting from an empty model.""")
            model_units = self.config['model'].get('model_units', {})
            self.copasi_model_object = new_model(
                name='CopasiProcess TimeCourseModel',
                **model_units
            )

        # handle context of species output
        context_type = self.config['species_context']
        self.species_context_key = f'floating_species_{context_type}'
        self.use_counts = 'concentrations' in context_type

        # Get a list of reactions
        self._set_reaction_changes()
        reactions = get_reactions(model=self.copasi_model_object)
        self.reaction_list = reactions.index.tolist() if reactions is not None else []
        # if not self.reaction_list:
        # raise AttributeError('No reactions could be parsed from this model. Your model must contain reactions to run.')

        # Get the species (floating only)  TODO: add boundary species
        self._set_species_changes()
        species_data = get_species(model=self.copasi_model_object)
        self.floating_species_list = species_data.index.tolist()
        self.floating_species_initial = species_data.particle_number.tolist() \
            if self.use_counts else species_data.concentration.tolist()

        # Get the list of parameters and their values (it is possible to run a model without any parameters)
        self._set_global_param_changes()
        model_parameters = get_parameters(model=self.copasi_model_object)
        self.model_parameters_list = model_parameters.index.tolist() \
            if isinstance(model_parameters, DataFrame) else []
        self.model_parameters_values = model_parameters.initial_value.tolist() \
            if isinstance(model_parameters, DataFrame) else []

        # Get a list of compartments
        self.compartments_list = get_compartments(model=self.copasi_model_object).index.tolist()

        # ----SOLVER: Get the solver (defaults to deterministic)
        self.method = self.config['method']

    def initial_state(self):
        # keep in mind that a valid simulation may not have global parameters
        model_parameters_dict = dict(
            zip(self.model_parameters_list, self.model_parameters_values))

        floating_species_dict = dict(
            zip(self.floating_species_list, self.floating_species_initial))

        return {
            'time': 0.0,
            'model_parameters': model_parameters_dict,
            self.species_context_key: floating_species_dict,
        }

    def inputs(self):
        # dependent on species context set in self.config
        floating_species_type = {
            species_id: {
                '_type': 'float',
                '_apply': 'set'}
            for species_id in self.floating_species_list
        }

        model_params_type = {
            param_id: {
                '_type': 'float',
                '_apply': 'set'}
            for param_id in self.model_parameters_list
        }

        reactions_type = {
            reaction_id: 'float'
            for reaction_id in self.reaction_list
        }

        return {
            'time': 'float',
            self.species_context_key: floating_species_type,
            'model_parameters': model_params_type,
            'reactions': reactions_type}

    def outputs(self):
        floating_species_type = {
            species_id: {
                '_type': 'float',
                '_apply': 'set'}
            for species_id in self.floating_species_list
        }
        return {
            'time': 'float',
            self.species_context_key: floating_species_type}

    def update(self, inputs, interval):
        # set copasi values according to what is passed in states for concentrations
        for cat_id, value in inputs[self.species_context_key].items():
            set_type = 'particle_number' if 'counts' in self.species_context_key else 'concentration'
            species_config = {
                'name': cat_id,
                'model': self.copasi_model_object,
                set_type: value}
            set_species(**species_config)

        # run model for "interval" length; we only want the state at the end
        timecourse = run_time_course(
            start_time=inputs['time'],
            duration=interval,
            update_model=True,
            model=self.copasi_model_object,
            method=self.method)

        # extract end values of concentrations from the model and set them in results
        results = {'time': interval}
        if self.use_counts:
            results[self.species_context_key] = {
                mol_id: float(get_species(
                    name=mol_id,
                    exact=True,
                    model=self.copasi_model_object
                ).particle_number[0])
                for mol_id in self.floating_species_list}
        else:
            results[self.species_context_key] = {
                mol_id: float(get_species(
                    name=mol_id,
                    exact=True,
                    model=self.copasi_model_object
                ).concentration[0])
                for mol_id in self.floating_species_list}

        return results

    def _set_reaction_changes(self):
        # ----REACTIONS: set reactions
        existing_reactions = get_reactions(model=self.copasi_model_object)
        existing_reaction_names = existing_reactions.index.tolist() if existing_reactions is not None else []
        reaction_changes = self.model_changes.get('reaction_changes', [])
        if reaction_changes:
            for reaction_change in reaction_changes:
                reaction_name: str = reaction_change['reaction_name']
                param_changes: list[dict[str, float]] = reaction_change['parameter_changes']
                scheme_change: str = reaction_change.get('reaction_scheme')
                # handle changes to existing reactions
                if param_changes:
                    for param_name, param_change_val in param_changes:
                        set_reaction_parameters(param_name, value=param_change_val, model=self.copasi_model_object)
                if scheme_change:
                    set_reaction(name=reaction_name, scheme=scheme_change, model=self.copasi_model_object)
                # handle new reactions
                if reaction_name not in existing_reaction_names and scheme_change:
                    add_reaction(reaction_name, scheme_change, model=self.copasi_model_object)

    def _set_species_changes(self):
        # ----SPECS: set species changes
        species_changes = self.model_changes.get('species_changes', [])
        if species_changes:
            for species_change in species_changes:
                if isinstance(species_change, dict):
                    species_name = species_change.pop('name')
                    changes_to_apply = {}
                    for spec_param_type, spec_param_value in species_change.items():
                        if spec_param_value:
                            changes_to_apply[spec_param_type] = spec_param_value
                    set_species(**changes_to_apply, model=self.copasi_model_object)

    def _set_global_param_changes(self):
        # ----GLOBAL PARAMS: set global parameter changes
        global_parameter_changes = self.model_changes.get('global_parameter_changes', [])
        if global_parameter_changes:
            for param_change in global_parameter_changes:
                param_name = param_change.pop('name')
                for param_type, param_value in param_change.items():
                    if not param_value:
                        param_change.pop(param_type)
                    # handle changes to existing params
                    set_parameters(name=param_name, **param_change, model=self.copasi_model_object)
                    # set new params
                    global_params = get_parameters(model=self.copasi_model_object)
                    if global_params:
                        existing_global_parameters = global_params.index
                        if param_name not in existing_global_parameters:
                            assert param_change.get('initial_concentration') is not None, "You must pass an initial_concentration value if adding a new global parameter."
                            add_parameter(name=param_name, **param_change, model=self.copasi_model_object)



import traceback
import subprocess
import sys
import importlib
from dataclasses import dataclass
from pprint import pformat
from types import ModuleType
from typing import Optional


def handle_exception(error_key: str | Exception = "bio-compose-error") -> str:
    tb_str = traceback.format_exc()
    if isinstance(error_key, Exception):
        error_key = str(error_key)
    error_message = pformat(f"{error_key}:\n{tb_str}")
    return error_message


def handle_sbml_exception() -> str:
    return handle_exception("time-course-simulation-error")


@dataclass
class SimulatorImport:
    simulator_id: str
    module: Optional[ModuleType] = None


def dynamic_simulator_import(module_name: str) -> SimulatorImport:
    """
    Import a single simulator module at the highest level. For use in async dynamic env creation.

    :param module_name: (`str`) The name of the module to import.
    :return: (`SimulatorImport`) The imported simulator.
    """
    module = None
    try:
        module = __import__(module_name)
    except ImportError as e:
        print(e)

    return SimulatorImport(simulator_id=module_name, module=module)


def dynamic_simulator_install(simulators: list[str], verbose: bool = True):
    """
    Dynamically installs required simulator libraries.

    :param simulators: (`list[str]`) list of simulator libraries to install
    :param verbose: (`bool`) whether to print progress confirmations
    """
    for sim in simulators:
        try:
            # Check if the simulator is already installed
            importlib.import_module(sim)
            print(f"{sim} is already installed.") if verbose else None
        except ImportError:
            # Install using pip in the current environment
            print(f"Installing {sim}...") if verbose else None
            subprocess.check_call([sys.executable, "-m", "pip", "install", sim])
            print(f"{sim} installed successfully.") if verbose else None


def new_document(name, address, _type, config, inputs, outputs, local_implementations=True, add_emitter=True):
    doc = {
        name: {
            "address": f"local:{address}" if local_implementations else address,
            "_type": _type,
            "config": config,
            "inputs": inputs,
            "outputs": outputs,
        }
    }

    if add_emitter:
        doc["emitter"] = {
            "address": f"local:ram-emitter",
            "_type": "step",
            "inputs": inputs
        }

    return doc



import os
import re
import zipfile as zf
import xml.etree.ElementTree as ET
from tempfile import mkdtemp
from typing import List, Optional, Dict
from pathlib import Path

import h5py
import numpy as np
import requests
import libsbml


class FilePath(Path):
    name: str

    def __init__(self, *args, name: str):
        super().__init__(*args)
        self.name = name


def fetch_biomodel_sbml_file(biomodel_id: str, save_dir: Optional[str] = None) -> FilePath:
    url = f'https://www.ebi.ac.uk/biomodels/search/download?models={biomodel_id}'
    headers = {'accept': '*/*'}
    response = requests.get(url, headers=headers)
    model_filename = f'{biomodel_id}.xml'
    dirpath = save_dir or mkdtemp()  # os.getcwd()
    response_zip_fp = os.path.join(dirpath, 'results.zip')
    if not os.path.exists(response_zip_fp):
        try:
            with open(response_zip_fp, 'wb') as f:
                f.write(response.content)

            with zf.ZipFile(response_zip_fp, 'r') as zipRef:
                zipRef.extract(model_filename, dirpath)

            os.remove(response_zip_fp)
            p = os.path.join(dirpath, model_filename)

            return FilePath(p, name=model_filename)
        except Exception as e:
            print(e)


def read_xyz(xyz_filepath: os.PathLike[str]) -> str:
    lines = []
    with open(xyz_filepath) as f:
        file_contents = f.read().splitlines()
        for i, line in enumerate(file_contents):
            if not file_contents[i] == file_contents[-1]:
                line = f"{line};"
            lines.append(line)
    return " ".join(lines)


def get_sbml_species_mapping(sbml_fp: str) -> dict:
    """

    Args:
        - sbml_fp: `str`: path to the SBML model file.

    Returns:
        Dictionary mapping of {sbml_species_names(usually the actual observable name): sbml_species_ids(ids used in the solver)}
    """
    # read file
    sbml_reader = libsbml.SBMLReader()
    sbml_doc = sbml_reader.readSBML(sbml_fp)
    sbml_model_object = sbml_doc.getModel()

    # parse and handle names/ids
    sbml_species_ids = []
    for spec in sbml_model_object.getListOfSpecies():
        spec_name = spec.name
        if not spec_name:
            spec.name = spec.getId()
        if not spec.name == "":
            sbml_species_ids.append(spec)
    names = list(map(lambda s: s.name, sbml_species_ids))
    species_ids = [spec.getId() for spec in sbml_species_ids]

    return dict(zip(names, species_ids))


def normalize_smoldyn_output_path_in_root(root_fp) -> str | None:
    new_path = None
    for root, dirs, files in os.walk(root_fp):
        for filename in files:
            if filename.endswith('out.txt'):
                original_path = os.path.join(root, filename)
                new_path = os.path.join(root, 'modelout.txt')
                os.rename(original_path, new_path)

    return new_path


def format_smoldyn_configuration(filename: str) -> None:
    config = read_smoldyn_simulation_configuration(filename)
    disable_smoldyn_graphics_in_simulation_configuration(configuration=config)
    return write_smoldyn_simulation_configuration(configuration=config, filename=filename)


def read_smoldyn_simulation_configuration(filename: str) -> List[str]:
    ''' Read a configuration for a Smoldyn simulation

    Args:
        filename (:obj:`str`): path to model file

    Returns:
        :obj:`list` of :obj:`str`: simulation configuration
    '''
    with open(filename, 'r') as file:
        return [line.strip('\n') for line in file]


def write_smoldyn_simulation_configuration(configuration: List[str], filename: str):
    ''' Write a configuration for Smoldyn simulation to a file

    Args:
        configuration
        filename (:obj:`str`): path to save configuration
    '''
    with open(filename, 'w') as file:
        for line in configuration:
            file.write(line)
            file.write('\n')


def disable_smoldyn_graphics_in_simulation_configuration(configuration: List[str]):
    ''' Turn off graphics in the configuration of a Smoldyn simulation

    Args:
        configuration (:obj:`list` of :obj:`str`): simulation configuration
    '''
    for i_line, line in enumerate(configuration):
        if line.startswith('graphics '):
            configuration[i_line] = re.sub(r'^graphics +[a-z_]+', 'graphics none', line)


def unpack_omex_archive(archive_filepath: str, working_dir: str, use_temp: bool = True, write: bool = False) -> str:
    archive_name = archive_filepath.replace('.omex', '')
    unpacking_dirpath = working_dir if not use_temp else mkdtemp()

    if not os.path.exists(unpacking_dirpath) and write:
        os.mkdir(unpacking_dirpath)

    with zf.ZipFile(archive_filepath, 'r') as f:
        f.extractall(path=unpacking_dirpath)

    return unpacking_dirpath


def get_archive_model_filepath(config_model_source: str) -> str:
    return [[os.path.join(root, f) for f in files if f.endswith('.xml') and not f.lower().startswith('manifest')][0] for
            root, _, files in os.walk(config_model_source)][0]


def get_sedml_time_config(sedml_filepath: str) -> Dict:
    # Parse the XML file
    tree = ET.parse(sedml_filepath)
    root = tree.getroot()

    namespace = {'sedml': 'http://sed-ml.org/sed-ml/level1/version3'}

    utc_elements = root.findall('.//sedml:uniformTimeCourse', namespace)

    # Extract data from each uniformTimeCourse element
    utc_data = []
    for utc in utc_elements:
        data = {
            'id': utc.get('id'),
            'initialTime': utc.get('initialTime'),
            'outputStartTime': utc.get('outputStartTime'),
            'outputEndTime': utc.get('outputEndTime'),
            'numberOfPoints': utc.get('numberOfPoints'),
            'algorithm': utc.find('.//sedml:algorithm', namespace).get('kisaoID')
        }
        utc_data.append(data)

    return utc_data[0]


def get_archive_file_location(archive_root: str, filename: str) -> FilePath:
    for f in os.listdir(archive_root):
        if f.endswith(filename):
            path = os.path.join(archive_root, f)
            print(path)
            return FilePath(name=filename, path=path)


def get_model_file_location(archive_root: str) -> FilePath:
    return get_archive_file_location(archive_root, '.xml')


def get_published_t(omex_dirpath: str = None, report_fp: str = None) -> np.ndarray:
    report_fp = get_archive_file_location(
        omex_dirpath, 'reports.h5') if not report_fp else report_fp

    published_outputs = read_report_outputs(report_fp)
    return published_outputs.data[0].data


def read_report_outputs(report_file_path) -> Dict:
    """Read the outputs from all species in the given report file from biosimulations output.
        Args:
            report_file_path (str): The path to the simulation.sedml/report.h5 HDF5 file.
    """
    # TODO: implement auto gen from run id here.
    outputs = []
    with h5py.File(report_file_path, 'r') as f:
        k = list(f.keys())
        group_path = k[0] + '/report'
        if group_path in f:
            group = f[group_path]
            dataset_labels = group.attrs['sedmlDataSetLabels']
            for label in dataset_labels:
                dataset_index = list(dataset_labels).index(label)
                data = group[()]
                specific_data = data[dataset_index]
                output = {"dataset_label": label, "data": specific_data}
                outputs.append(output)

            return {"report_path": report_file_path, "data": outputs}
        else:
            print(f"Group '{group_path}' not found in the file.")


import os
import re
import uuid
from abc import abstractmethod
from logging import warn
from uuid import uuid4
from typing import *

from process_bigraph import Step, Process
from process_bigraph.composite import Emitter
from pymongo import ASCENDING, MongoClient
from pymongo.database import Database
from simulariumio import InputFileData, UnitData, DisplayData, DISPLAY_TYPE
from simulariumio.smoldyn import SmoldynData

from bsp.data_generators import SBML_EXECUTORS
from bsp.io import get_sbml_species_mapping
from bsp.utils.simularium_utils import translate_data_object, write_simularium_file, calculate_agent_radius

VERBOSE = False
try:
    import smoldyn as sm
    from smoldyn._smoldyn import MolecState
except:
    if VERBOSE:
        print(
            '\nPLEASE NOTE: Smoldyn is not correctly installed on your system which prevents you from ' 
            'using the SmoldynProcess. Please refer to the README for further information '
            'on installing Smoldyn.'
        )

HISTORY_INDEXES = [
    'data.time',
    [('experiment_id', ASCENDING),
     ('data.time', ASCENDING),
     ('_id', ASCENDING)],
]
CONFIGURATION_INDEXES = [
    'experiment_id',
]
SECRETS_PATH = 'secrets.json'


# -- emitters -- #

class MongoDatabaseEmitter(Emitter):
    client_dict: Dict[int, MongoClient] = {}
    config_schema = {
        'connection_uri': 'string',
        'experiment_id': 'maybe[string]',
        'emit_limit': {
            '_type': 'integer',
            '_default': 4000000
        },
        'database': 'maybe[string]'
    }

    @classmethod
    def create_indexes(cls, table: Any, columns: List[Any]) -> None:
        """Create the listed column indexes for the given DB table."""
        for column in columns:
            table.create_index(column)

    def __init__(self, config, core) -> None:
        """Config may have 'host' and 'database' items. The config passed is expected to be:

                {'experiment_id':,
                 'emit_limit':,
                 'embed_path':}

                TODO: Automate this process for the user in builder
        """
        super().__init__(config)
        self.core = core
        self.experiment_id = self.config.get('experiment_id', str(uuid.uuid4()))
        # In the worst case, `breakdown_data` can underestimate the size of
        # data by a factor of 4: len(str(0)) == 1 but 0 is a 4-byte int.
        # Use 4 MB as the breakdown limit to stay under MongoDB's 16 MB limit.
        self.emit_limit = self.config['emit_limit']

        # create new MongoClient per OS process
        curr_pid = os.getpid()
        if curr_pid not in MongoDatabaseEmitter.client_dict:
            MongoDatabaseEmitter.client_dict[curr_pid] = MongoClient(
                config['connection_uri'])
        self.client: MongoClient = MongoDatabaseEmitter.client_dict[curr_pid]

        # extract objects from current mongo client instance
        self.db: Database = getattr(self.client, self.config.get('database', 'simulations'))
        self.history_collection: Collection = getattr(self.db, 'history')
        self.configuration: Collection = getattr(self.db, 'configuration')

        # create column indexes for the given collection objects
        self.create_indexes(self.history_collection, HISTORY_INDEXES)
        self.create_indexes(self.configuration, CONFIGURATION_INDEXES)

        self.fallback_serializer = make_fallback_serializer_function(self.core)

    def query(self, query):
        return self.history_collection.find_one(query)

    def history(self):
        return [v for v in self.history_collection.find()]

    def flush_history(self):
        for v in self.history():
            self.history_collection.delete_one(v)

    def update(self, inputs):
        self.history_collection.insert_one(inputs)
        return {}


# -- simulators -- #

class SmoldynStep(Step):
    config_schema = {
        'model_filepath': 'string',
        'animate': {
            '_type': 'boolean',
            '_default': False
        },
        'duration': 'maybe[integer]',
        'dt': 'maybe[float]',
        'initial_species_counts': 'maybe[tree[float]]',
        'initial_mol_position': 'maybe[list[float]]',  # of particles/molecules
        'initial_mol_state': 'maybe[integer]',

        # TODO: Add a more nuanced way to describe and configure dynamic difcs given species interaction patterns
    }

    def __init__(self, config, core):
        """A new instance of `SmoldynProcess` based on the `config` that is passed. The schema for the config to be passed in
            this object's constructor is as follows:

            config_schema = {
                'model_filepath': 'string',  <-- analogous to python `str`
                'animate': 'bool'  <-- of type `bigraph_schema.base_types.bool`

            # TODO: It would be nice to have classes associated with this.
        """
        super().__init__(config=config, core=core)

        # specify the model fp for clarity
        self.model_filepath = self.config.get('model_filepath')

        # enforce model filepath passing
        if not self.model_filepath:
            raise ValueError(
                '''
                    The Process configuration requires a Smoldyn model filepath to be passed.
                    Please specify a 'model_filepath' in your instance configuration.
                '''
            )

        # initialize the simulator from a Smoldyn MinE.txt file.
        self.simulation: sm.Simulation = sm.Simulation.fromFile(self.model_filepath)

        # set default starting position of molecules/particles (assume all)
        self.initial_mol_position = self.config.get('initial_mol_position', [0.0, 0.0, 0.0])
        self.initial_mol_state = self.config.get('initial_mol_state', 0)

        # get a list of the simulation species
        species_count = self.simulation.count()['species']
        counts = self.config.get('initial_species_counts')
        self.initial_species_counts = counts
        self.species_names: List[str] = []
        for index in range(species_count):
            species_name = self.simulation.getSpeciesName(index)
            if 'empty' not in species_name.lower():
                self.species_names.append(species_name)

        self.initial_species_state = {}
        self.initial_mol_state = {}
        initial_mol_counts = {spec_name: self.simulation.getMoleculeCount(spec_name, MolecState.all) for spec_name in self.species_names}
        for species_name, count in initial_mol_counts.items():
            self.initial_species_state[species_name] = count
            for _ in range(count):
                self.initial_mol_state[str(uuid4())] = {
                    'coordinates': self.initial_mol_position,
                    'species_id': species_name,
                    'state': self.initial_mol_state
                }

        # sort for logistical mapping to species names (i.e: ['a', 'b', c'] == ['0', '1', '2']
        self.species_names.sort()

        # make species counts of molecules dataset for output
        self.simulation.addOutputData('species_counts')
        # write molcounts to counts dataset at every timestep (shape=(n_timesteps, 1+n_species <-- one for time)): [timestep, countSpec1, countSpec2, ...]
        self.simulation.addCommand(cmd='molcount species_counts', cmd_type='E')

        # make molecules dataset (molecule information) for output
        self.simulation.addOutputData('molecules')
        # write coords to dataset at every timestep (shape=(n_output_molecules, 7)): seven being [timestep, smol_id(species), mol_state, x, y, z, mol_serial_num]
        self.simulation.addCommand(cmd='listmols molecules', cmd_type='E')

        # initialize the molecule ids based on the species names. We need this value to properly emit the schema, which expects a single value from this to be a str(int)
        # the format for molecule_ids is expected to be: 'speciesId_moleculeNumber'
        self.molecule_ids = list(self.initial_mol_state.keys())

        # get the simulation boundaries, which in the case of Smoldyn denote the physical boundaries
        # TODO: add a verification method to ensure that the boundaries do not change on the next step...
        self.boundaries: Dict[str, List[float]] = dict(zip(['low', 'high'], self.simulation.getBoundaries()))

        # create a re-usable counts and molecules type to be used by both inputs and outputs
        self.counts_type = {
            species_name: 'integer'
            for species_name in self.species_names
        }

        self.output_port_schema = {
            'species_counts': {
                species_name: 'integer'
                for species_name in self.species_names
            },
            'molecules': 'tree[string]',  # self.molecules_type
            'results_file': 'string'
        }

        # set time if applicable
        self.duration = self.config.get('duration')
        self.dt = self.config.get('dt', self.simulation.dt)

        # set graphics (defaults to False)
        if self.config['animate']:
            self.simulation.addGraphics('opengl_better')

        self._specs = [None for _ in self.species_names]
        self._vals = dict(zip(self.species_names, [[] for _ in self.species_names]))

    # def initial_state(self):
    #     return {
    #         'species_counts': self.initial_species_state,
    #         'molecules': self.initial_mol_state
    #     }

    def inputs(self):
        # schema = self.output_port_schema.copy()
        # schema.pop('results_file')
        # return schema
        return {}

    def outputs(self):
        return self.output_port_schema

    def update(self, inputs) -> Dict:
        # reset the molecules, distribute the mols according to self.boundarieså
        # for name in self.species_names:
        #     self.set_uniform(
        #         species_name=name,
        #         count=inputs['species_counts'][name],
        #         kill_mol=False
        #     )

        # run the simulation for a given interval if specified, otherwise use builtin time
        if self.duration is not None:
            self.simulation.run(stop=self.duration, dt=self.simulation.dt, overwrite=True)
        else:
            self.simulation.runSim()

        # get the counts data, clear the buffer
        counts_data = self.simulation.getOutputData('species_counts')

        # get the final counts for the update
        final_count = counts_data[-1]
        # remove the timestep from the list
        final_count.pop(0)

        # get the data based on the commands added in the constructor, clear the buffer
        molecules_data = self.simulation.getOutputData('molecules')

        # create an empty simulation state mirroring that which is specified in the schema
        simulation_state = {
            'species_counts': {},
            'molecules': {}
        }

        # get and populate the species counts
        for index, name in enumerate(self.species_names):
            simulation_state['species_counts'][name] = counts_data[index]
            # input_counts = simulatio['species_counts'][name]
            # simulation_state['species_counts'][name] = int(final_count[index]) - input_counts

        # clear the list of known molecule ids and update the list of known molecule ids (convert to an intstring)
        # self.molecule_ids.clear()
        # for molecule in molecules_data:
            # self.molecule_ids.append(str(uuid4()))

        # get and populate the output molecules
        for i, single_mol_data in enumerate(molecules_data):
            mol_species_index = int(single_mol_data[1]) - 1
            mol_id = str(uuid4())
            simulation_state['molecules'][mol_id] = {
                'coordinates': single_mol_data[3:6],
                'species_id': self.species_names[mol_species_index],
                'state': str(int(single_mol_data[2]))
            }

        # mols = []
        # for index, mol_id in enumerate(self.molecule_ids):
        #     single_molecule_data = molecules_data[index]
        #     single_molecule_species_index = int(single_molecule_data[1]) - 1
        #     mols.append(single_molecule_species_index)
        #     simulation_state['molecules'][mol_id] = {
        #         'coordinates': single_molecule_data[3:6],
        #         'species_id': self.species_names[single_molecule_species_index],
        #         'state': str(int(single_molecule_data[2]))
        #     }

        # TODO -- post processing to get effective rates

        # TODO: adjust this for a more dynamic dir struct
        model_dir = os.path.dirname(self.model_filepath)
        for f in os.listdir(model_dir):
            if f.endswith('.txt') and 'out' in f:
                simulation_state['results_file'] = os.path.join(model_dir, f)

        return simulation_state

    def set_uniform(
            self,
            species_name: str,
            count: int,
            kill_mol: bool = True
    ) -> None:
        """Add a distribution of molecules to the solution in
            the simulation memory given a higher and lower bound x,y coordinate. Smoldyn assumes
            a global boundary versus individual species boundaries. Kills the molecule before dist if true.

            TODO: If pymunk expands the species compartment, account for
                  expanding `highpos` and `lowpos`. This method should be used within the body/logic of
                  the `update` class method.

            Args:
                species_name:`str`: name of the given molecule.
                count:`int`: number of molecules of the given `species_name` to add.
                kill_mol:`bool`: kills the molecule based on the `name` argument, which effectively
                    removes the molecule from simulation memory.
        """
        # kill the mol, effectively resetting it
        if kill_mol:
            self.simulation.runCommand(f'killmol {species_name}')

        # TODO: eventually allow for an expanding boundary ie in the configuration parameters (pymunk?), which is defies the methodology of smoldyn

        # redistribute the molecule according to the bounds
        self.simulation.addSolutionMolecules(
            species=species_name,
            number=count,
            highpos=self.boundaries['high'],
            lowpos=self.boundaries['low']
        )


class SimulariumSmoldynStep(Step):
    """
        agent_data should have the following structure:

        {species_name(type):
            {display_type: DISPLAY_TYPE.<REQUIRED SHAPE>,
             (mass: `float` AND density: `float`) OR (radius: `float`)

    """
    config_schema = {
        'output_dest': 'string',
        'box_size': 'float',  # as per simulariumio
        'spatial_units': {
            '_default': 'nm',
            '_type': 'string'
        },
        'temporal_units': {
            '_default': 'ns',
            '_type': 'string'
        },
        'translate_output': {
            '_default': True,
            '_type': 'boolean'
        },
        'write_json': {
            '_default': True,
            '_type': 'boolean'
        },
        'run_validation': {
            '_default': True,
            '_type': 'boolean'
        },
        'file_save_name': 'maybe[string]',
        'translation_magnitude': 'maybe[float]',
        'meta_data': 'maybe[tree[string]]',
        'agent_display_parameters': 'maybe[tree[string]]'  # as per biosim simularium
    }

    def __init__(self, config, core):
        super().__init__(config=config, core=core)

        # io params
        self.output_dest = self.config['output_dest']
        self.write_json = self.config['write_json']
        self.filename = self.config.get('file_save_name')

        # display params
        self.box_size = self.config['box_size']
        self.translate_output = self.config['translate_output']
        self.translation_magnitude = self.config.get('translation_magnitude')
        self.agent_display_parameters = self.config.get('agent_display_parameters', {})

        # units params
        self.spatial_units = self.config['spatial_units']
        self.temporal_units = self.config['temporal_units']

        # info params
        self.meta_data = self.config.get('meta_data')
        self.run_validation = self.config['run_validation']

    def inputs(self):
        return {'results_file': 'string', 'species_names': 'list[string]'}

    def outputs(self):
        return {'simularium_file': 'string'}

    def update(self, inputs):
        # get job params
        in_file = inputs['results_file']
        file_data = InputFileData(in_file)

        # get species data for display data
        species_names = inputs['species_names']

        # generate simulariumio Smoldyn Data TODO: should display data be gen for each species type or n number of instances of that type?
        display_data = self._generate_display_data(species_names)
        io_data = SmoldynData(
            smoldyn_file=file_data,
            spatial_units=UnitData(self.spatial_units),
            time_units=UnitData(self.temporal_units),
            display_data=display_data,
            meta_data=self.meta_data,
            center=True
        )

        # translate reflections if needed
        if self.translate_output:
            io_data = translate_data_object(data=io_data, box_size=self.box_size, translation_magnitude=self.translation_magnitude)
        # write data to simularium file
        if self.filename is None:
            self.filename = in_file.split('/')[-1].replace('.', '') + "-simulation"

        save_path = os.path.join(self.output_dest, self.filename)
        write_simularium_file(data=io_data, simularium_fp=save_path, json=self.write_json, validate=self.run_validation)
        result = {'simularium_file': save_path + '.simularium'}

        return result

    def _generate_display_data(self, species_names) -> Dict | None:
        # user is specifying display data for agents
        if isinstance(self.agent_display_parameters, dict) and len(self.agent_display_parameters.keys()) > 0:
            display_data = {}
            for name in species_names:
                display_params = self.agent_display_parameters[name]

                # handle agent radius
                radius_param = display_params.get('radius')

                # user has passed a mass and density for a given agent
                if radius_param is None:
                    radius_param = calculate_agent_radius(m=display_params['mass'], rho=display_params['density'])

                # make kwargs for display data
                display_data_kwargs = {
                    'name': name,
                    'display_type': DISPLAY_TYPE[display_params['display_type']],
                    'radius': radius_param
                }

                # check if self.agent_params as been passed as a mapping of species_name: {species_mass: , species_shape: }
                display_data[name] = DisplayData(**display_data_kwargs)

            return display_data

        return None


# -- utils --

def generate_simularium_file(
        input_fp: str,
        dest_dir: str,
        box_size: float,
        translate_output: bool = True,
        write_json: bool = True,
        run_validation: bool = True,
        agent_parameters: Dict[str, Dict[str, Any]] = None
) -> Dict[str, str]:
    species_names = []
    float_pattern = re.compile(r'^-?\d+(\.\d+)?$')
    with open(input_fp, 'r') as f:
        output = [l.strip() for l in f.readlines()]
        for line in output:
            datum = line.split(' ')[0]
            # Check if the datum is not a float string
            if not float_pattern.match(datum):
                species_names.append(datum)
        f.close()
    species_names = list(set(species_names))

    simularium = SimulariumSmoldynStep(config={
        'output_dest': dest_dir,
        'box_size': box_size,
        'translate_output': translate_output,
        'file_save_name': None,
        'write_json': write_json,
        'run_validation': run_validation,
        'agent_display_parameters': agent_parameters,
    })

    return simularium.update(inputs={
        'results_file': input_fp,
        'species_names': species_names
    })





# -- Output data generators: -- #

class OutputGenerator(Step):
    config_schema = {
        'input_file': 'string',
        'context': 'string',
    }

    def __init__(self, config, core):
        super().__init__(config, core)
        self.input_file = self.config['input_file']
        self.context = self.config.get('context')
        if self.context is None:
            raise ValueError("context (i.e., simulator name) must be specified in this processes' config.")

    @abstractmethod
    def generate(self, parameters: Optional[Dict[str, Any]] = None):
        """Abstract method for generating output data upon which to base analysis from based on its origin.

        This can be used for logic of any scope.
        NOTE: args and kwargs are not defined in this function, but rather should be defined by the
        inheriting class' constructor: i,e; start_time, etc.

        Kwargs relate only to the given simulator api you are working with.
        """
        pass

    def initial_state(self):
        # base class method
        return {
            'output_data': {}
        }

    def inputs(self):
        return {
            'parameters': 'tree[any]'
        }

    def outputs(self):
        return {
            'output_data': 'tree[any]'
        }

    def update(self, state):
        parameters = state.get('parameters') if isinstance(state, dict) else {}
        data = self.generate(parameters)
        return {'output_data': data}


class TimeCourseOutputGenerator(OutputGenerator):
    # NOTE: we include defaults here as opposed to constructor for the purpose of deliberate declaration within .json state representation.
    config_schema = {
        # 'input_file': 'string',
        # 'context': 'string',
        'start_time': {
            '_type': 'integer',
            '_default': 0
        },
        'end_time': {
            '_type': 'integer',
            '_default': 10
        },
        'num_steps': {
            '_type': 'integer',
            '_default': 100
        },
    }

    def __init__(self, config, core):
        super().__init__(config, core)
        if not self.input_file.endswith('.xml'):
            raise ValueError('Input file must be a valid SBML (XML) file')

        self.start_time = self.config.get('start_time')
        self.end_time = self.config.get('end_time')
        self.num_steps = self.config.get('num_steps')
        self.species_mapping = get_sbml_species_mapping(self.input_file)

    def initial_state(self):
        # TODO: implement this
        pass

    def generate(self, parameters: Optional[Dict[str, Any]] = None):
        # TODO: add kwargs (initial state specs) here
        executor = SBML_EXECUTORS[self.context]
        data = executor(self.input_file, self.start_time, self.end_time, self.num_steps)

        return data





"""
Data Model relating to Bigraph (process-bigraph) specification: nodes, composites, etc.
"""


import os
from dataclasses import dataclass, Field
from typing import List, Dict, Any


class NodeSpec(dict):
    def __init__(self, _type: str, address: str, config: Dict[str, Any], inputs: Dict[str, List[str]], outputs: Dict[str, List[str]], name: str = None):
        super().__init__()
        self._type = _type
        self.address = address
        self.config = config
        self.inputs = inputs
        self.outputs = outputs
        self.name = name


def node_spec(_type: str, address: str, config: Dict[str, Any], inputs: Dict[str, List[str]], outputs: Dict[str, List[str]], name: str = None) -> Dict[str, Any]:
    spec = {
        '_type': _type,
        'address': address,
        'config': config,
        'inputs': inputs,
        'outputs': outputs
    }

    return {name: spec} if name else spec


def step_node_spec(address: str, config: Dict[str, Any], inputs: Dict[str, Any], outputs: Dict[str, Any], name: str = None):
    return node_spec(name=name, _type="step", address=address, config=config, inputs=inputs, outputs=outputs)


def process_node_spec(address: str, config: Dict[str, Any], inputs: Dict[str, Any], outputs: Dict[str, Any], name: str = None):
    return node_spec(name=name, _type="process", address=address, config=config, inputs=inputs, outputs=outputs)


def time_course_node_spec(input_file: str, context: str, start_time: int, end_time: int, num_steps: int):
    config = {
        'input_file': input_file,
        'start_time': start_time,
        'end_time': end_time,
        'num_steps': num_steps,
        'context': context
    }
    return step_node_spec(
        address='local:time-course-output-generator',
        config=config,
        inputs={
            'parameters': [f'parameters_store_{context}']
        },
        outputs={
            'output_data': [f'output_data_store_{context}']
        }
    )



import os
import numpy as np
import matplotlib.pyplot as plt
import imageio.v2 as imageio
import io
import base64

from IPython.core.display import HTML
from IPython.core.display_functions import display
from IPython.display import display, HTML
from imageio import v2 as imageio
from matplotlib import pyplot as plt


def sort_results(results):
    emitter_results = results[('emitter',)]
    sorted_results = {'fields': {
        key: [] for key in emitter_results[0]['fields'].keys()
    }, 'time': []}

    for results in emitter_results:
        time = results['global_time']
        fields = results['fields']
        sorted_results['time'].append(time)
        for key, value in fields.items():
            sorted_results['fields'][key].append(value)
    return sorted_results


def plot_time_series(
        results,
        field_names=None,
        coordinates=None,
        out_dir=None,
        filename='time_series.png',
        display=False,
):
    """
    Plots time series for specified fields and coordinates from the results dictionary.

    :param results: Dictionary containing the results with keys 'time' and 'fields'.
                    Example: {'time': [0.1, 0.2, 0.3 ...], 'fields': {'glucose': [nparray_time0, nparray_time1], ...}}
    :param field_names: List of field names to plot.
                        Example: ['glucose', 'other chemical']
    :param coordinates: List of coordinates (indices) to plot.
                        Example: [(0, 0), (1, 2)]
    """
    coordinates = coordinates or [(0, 0)]
    field_names = field_names or ['glucose', 'acetate', 'biomass']
    sorted_results = sort_results(results)
    time = sorted_results['time']

    # Initialize the plot
    fig, ax = plt.subplots(figsize=(12, 6))

    for field_name in field_names:
        if field_name in sorted_results['fields']:
            field_data = sorted_results['fields'][field_name]

            for coord in coordinates:
                x, y = coord
                time_series = [field_data[t][x, y] for t in range(len(time))]
                ax.plot(time, time_series, label=f'{field_name} at {coord}')
                # plot log scale on y axis
                # ax.set_yscale('log')
        else:
            print(f"Field '{field_name}' not found in results['fields']")

    # Adding plot labels and legend
    ax.set_xlabel('Time')
    ax.set_ylabel('Value')
    ax.set_title('Time Series Plot')
    ax.legend()

    # Save the plot to a file
    if out_dir is not None:
        os.makedirs(out_dir, exist_ok=True)
        filepath = os.path.join(out_dir, filename)
        print(f'saving {filepath}')
        plt.savefig(filepath)
    else:
        filepath = filename
    if display:
        # Display the plot
        plt.show()


def plot_snapshots(
        results,
        field_names,
        times
):
    """
    Plots snapshots of specified fields at specified times from the results dictionary.

    :param results: Dictionary containing the results with keys 'time' and 'fields'.
                    Example: {'time': [0.1, 0.2, 0.3 ...], 'fields': {'glucose': [nparray_time0, nparray_time1], ...}}
    :param field_names: List of field names to plot.
                        Example: ['glucose', 'other chemical']
    :param times: List of times at which to plot the snapshots.
                  Example: [0.1, 0.3]
    """
    sorted_results = sort_results(results)
    time_indices = [sorted_results['time'].index(t) for t in times]

    num_rows = len(times)
    num_cols = len(field_names)

    # Compute global min and max for each field
    global_min_max = {}
    for field_name in field_names:
        field_data_all_times = np.concatenate(
            [sorted_results['fields'][field_name][t].flatten() for t in range(len(sorted_results['time']))])
        global_min_max[field_name] = (np.min(field_data_all_times), np.max(field_data_all_times))

    fig, axes = plt.subplots(num_rows, num_cols, figsize=(5 * num_cols, 5 * num_rows))

    for row, time_index in enumerate(time_indices):
        for col, field_name in enumerate(field_names):
            if field_name in sorted_results['fields']:
                field_data = sorted_results['fields'][field_name][time_index]
                ax = axes[row, col] if num_rows > 1 and num_cols > 1 else (axes[row] if num_rows > 1 else axes[col])
                im = ax.imshow(field_data, cmap='viridis', aspect='auto',
                               vmin=global_min_max[field_name][0], vmax=global_min_max[field_name][1])
                ax.set_title(f'{field_name} at t={sorted_results["time"][time_index]}')
                fig.colorbar(im, ax=ax)
            else:
                print(f"Field '{field_name}' not found in results['fields']")

    plt.tight_layout()
    plt.show()


def plot_species_distributions_to_gif(
        results,
        out_dir=None,
        filename='species_distribution.gif',
        title='',
        skip_frames=1
):
    # Sort the results as before
    sorted_results = sort_results(results)
    species_names = [key for key in sorted_results['fields'].keys()]
    n_species = len(species_names)
    times = sorted_results['time']
    n_times = len(times)

    # Compute global min and max for each species
    global_min_max = {
        species: (np.min(np.concatenate([sorted_results['fields'][species][i].flatten() for i in range(n_times)])),
                  np.max(np.concatenate([sorted_results['fields'][species][i].flatten() for i in range(n_times)])))
        for species in species_names}

    images = []
    for i in range(0, n_times, skip_frames):
        fig, axs = plt.subplots(1, n_species, figsize=(5 * n_species, 4))
        if n_species == 1:
            axs = [axs]

        for j, species in enumerate(species_names):
            ax = axs[j]
            vmin, vmax = global_min_max[species]
            img = ax.imshow(sorted_results['fields'][species][i], interpolation='nearest', vmin=vmin, vmax=vmax)
            ax.set_title(f'{species} at t = {times[i]:.2f}')
            plt.colorbar(img, ax=ax)

        fig.suptitle(title, fontsize=16)
        plt.subplots_adjust(wspace=0.05, hspace=0.2)
        plt.tight_layout(pad=0.2)

        # Save the current figure to a temporary buffer
        buf = io.BytesIO()
        plt.savefig(buf, format='png', dpi=120)
        buf.seek(0)
        images.append(imageio.imread(buf))
        buf.close()
        plt.close(fig)

    # Create the output directory if not exists
    if out_dir is not None:
        os.makedirs(out_dir, exist_ok=True)
        filepath = os.path.join(out_dir, filename)
    else:
        filepath = filename

    # Create and save the GIF with loop=0 for infinite loop
    print(f'saving {filepath}')
    imageio.mimsave(filepath, images, duration=0.5, loop=0)

    # Optionally display the GIF in a Jupyter notebook
    with open(filepath, 'rb') as file:
        data = file.read()
        data_url = 'data:image/gif;base64,' + base64.b64encode(data).decode()
    display(HTML(f'<img src="{data_url}" alt="{title}" style="max-width:100%;"/>'))


def plot_species_distributions_with_particles_to_gif(
        results,
        out_dir=None,
        filename='species_distribution_with_particles.gif',
        title='',
        skip_frames=1,
        bounds=(1.0, 1.0),
):
    # Sort the results
    sorted_results = sort_results(results)
    species_names = [key for key in sorted_results['fields'].keys()]
    n_species = len(species_names)
    times = sorted_results['time']
    n_times = len(times)
    # Compute global min and max for each species
    global_min_max = {
        species: (np.min(np.concatenate([sorted_results['fields'][species][i].flatten() for i in range(n_times)])),
                  np.max(np.concatenate([sorted_results['fields'][species][i].flatten() for i in range(n_times)])))
        for species in species_names}

    xmax, ymax = bounds  # unpack the bounds

    emitter_results = results[('emitter',)]

    images = []
    for i in range(0, n_times, skip_frames):
        fields = emitter_results[i]['fields']
        particles = emitter_results[i]['particles']

        fig, axs = plt.subplots(1, n_species, figsize=(5 * n_species, 4))
        if n_species == 1:
            axs = [axs]

        for j, species in enumerate(species_names):
            ax = axs[j]
            vmin, vmax = global_min_max[species]
            # Plot species distribution as heatmap
            flipped_field = fields[species][::-1, :]  # [::-1, :] flips the array along the first axis (rows), while keeping the second axis (columns)
            img = ax.imshow(flipped_field, interpolation='nearest', vmin=vmin, vmax=vmax,
                            extent=[0, xmax, 0, ymax])  # extent stretches the field

            ax.set_title(f'{species} at t = {times[i]:.2f}')
            plt.colorbar(img, ax=ax)

            for particle in particles:
                ax.scatter(particle['position'][0], particle['position'][1],
                           s=particle['size'],
                           color=particle['color']
                           )

        fig.suptitle(title, fontsize=16)
        plt.subplots_adjust(wspace=0.05, hspace=0.2)
        plt.tight_layout(pad=0.2)

        # Save the current figure to a temporary buffer
        buf = io.BytesIO()
        plt.savefig(buf, format='png', dpi=120)
        buf.seek(0)
        images.append(imageio.imread(buf))
        buf.close()
        plt.close(fig)

    # Create the output directory if it doesn't exist
    if out_dir is not None:
        os.makedirs(out_dir, exist_ok=True)
        filepath = os.path.join(out_dir, filename)
    else:
        filepath = filename

    # Create and save the GIF with loop=0 for infinite loop
    print(f'saving {filepath}')
    imageio.mimsave(filepath, images, duration=0.5, loop=0)

    # Optionally display the GIF in a Jupyter notebook
    with open(filepath, 'rb') as file:
        data = file.read()
        data_url = 'data:image/gif;base64,' + base64.b64encode(data).decode()
    display(HTML(f'<img src="{data_url}" alt="{title}" style="max-width:100%;"/>'))


def plot_particles(
        # total_time,
        env_size,
        history,
        out_dir=None,
        filename='multi_species_diffusion.gif',
        fps=20
):
    """
    Plot particle movements and save the animation as a GIF.

    Parameters:
    - total_time: Total time (frames) of the simulation.
    - env_size: Tuple indicating the xlim and ylim of the environment, as ((xmin, xmax), (ymin, ymax)).
    - history: History of particles at each time step.
    - filename: Filename for the output GIF.
    - fps: Frames per second for the output GIF.
    """
    fig, ax = plt.subplots()
    ax.set_xlim(*env_size[0])
    ax.set_ylim(*env_size[1])
    ax.set_aspect('equal')
    n_frames = len(history)

    images = []
    for frame in range(n_frames):  # Include initial position
        ax.clear()
        ax.set_title(f'time {frame}')
        ax.set_xlim(*env_size[0])
        ax.set_ylim(*env_size[1])
        ax.set_aspect('equal')

        particles = history[frame]
        for particle in particles:
            ax.scatter(particle['position'][0], particle['position'][1],
                       s=particle['size'], color=particle['color'])

        # Save the current figure to a temporary buffer
        buf = io.BytesIO()
        plt.savefig(buf, format='png', dpi=120)
        buf.seek(0)
        images.append(imageio.imread(buf))
        buf.close()

    plt.close(fig)

    # Create the output directory if not exists
    if out_dir is not None:
        os.makedirs(out_dir, exist_ok=True)
        filepath = os.path.join(out_dir, filename)
    else:
        filepath = filename

    # Create and save the GIF with loop=0 for infinite loop
    print(f'saving {filepath}')
    imageio.mimsave(filepath, images, fps=fps, loop=0)

    # Optionally display the GIF in a Jupyter notebook
    with open(filepath, 'rb') as file:
        data = file.read()
        data_url = 'data:image/gif;base64,' + base64.b64encode(data).decode()
    display(HTML(f'<img src="{data_url}" alt="Particle Diffusion" style="max-width:100%;"/>'))


"""
Dynamic FBA simulation
======================

Process for a pluggable dFBA simulation.
"""
import logging
import os
from pathlib import Path

import numpy as np
import warnings

import cobra
from cobra.io import load_model
from process_bigraph import Process, Composite

from bsp.data_model.sed import SedModel
from bsp.schemas.config import TimeCourseConfig
from bsp.viz.plot import plot_time_series, plot_species_distributions_to_gif


# Suppress warnings
logger = logging.getLogger('cobra')
logger.setLevel(logging.ERROR)
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=UserWarning, module="cobra.util.solver")
warnings.filterwarnings("ignore", category=FutureWarning, module="cobra.medium.boundary_types")


# TODO -- can set lower and upper bounds by config instead of hardcoding
MODEL_FOR_TESTING = load_model('textbook')
# MODEL_FOR_TESTING.reactions.EX_o2_e.lower_bound = -2  # Limiting oxygen uptake
# MODEL_FOR_TESTING.reactions.ATPM.lower_bound = 1     # Setting lower bound for ATP maintenance
# MODEL_FOR_TESTING.reactions.ATPM.upper_bound = 1     # Setting upper bound for ATP maintenance


class CobraProcess(Process):
    """FBA component of the dfba hybrid using COBRA."""
    config_schema = {
        'model': SedModel,
        'objective': {
            'domain': 'string',  # either protein or mrna
            'name': 'string',  # specific to the model: i.e., LacI
            'scaling_factor': 'float'
        }
    }

    def __init__(self, config=None, core=None):
        super().__init__(config, core)

        # create model
        model_file = self.config['model']['model_source']
        data_dir = Path(os.path.dirname(model_file))
        path = data_dir / model_file.split('/')[-1]
        self.model = load_model(str(path.resolve()))  # read_sbml_model(str(path.resolve()))

        # parse objective
        self.objective_domain = self.config['objective']['domain']
        self.objective_name = self.config['objective']['name']
        self.scaling_factor = self.config['objective'].get('scaling_factor', 10)

        # set objectives
        self.model.objective = {
            self.model.reactions.get_by_id(reaction.id): np.random.random()  # TODO: make this more realistic
            for reaction in self.model.reactions
        }

        # set even bounds
        for reaction in self.model.reactions:
            rand_bound = np.random.random()
            self.model.reactions.get_by_id(reaction.id).lower_bound = -rand_bound  # TODO: What to do here?
            self.model.reactions.get_by_id(reaction.id).upper_bound = rand_bound

    def initial_state(self):
        initial_fluxes = {}
        initial_solution = self.model.optimize()
        if initial_solution.status == 'optimal':
            initial_fluxes = {
                reaction.name: reaction.flux
                for reaction in self.model.reactions
            }

        return {'fluxes': initial_fluxes}

    def inputs(self):
        return {'reaction_fluxes': 'tree[float]'}

    def outputs(self):
        return {'fluxes': 'tree[float]'}

    def update(self, state, interval):
        for reaction_name, reaction_flux in state['reaction_fluxes'].items():
            for reaction in self.model.reactions:
                if reaction.name == reaction_name:
                    # 1. reset objective weights according to reaction fluxes directly
                    self.model.objective = {
                        self.model.reactions.get_by_id(reaction.id): reaction_flux
                    }

                    # 2. set lower bound with scaling factor and reaction fluxes
                    self.model.reactions.get_by_id(reaction.id).lower_bound = -self.scaling_factor * abs(reaction_flux)  # / (5 + abs(reaction_flux))

        # 3. solve for fluxes
        output_state = {}
        solution = self.model.optimize()
        if solution.status == "optimal":
            data = dict(zip(
                list(state['reaction_fluxes'].keys()),
                list(solution.fluxes.to_dict().values())
            ))
            output_state['fluxes'] = data

            # TODO: do we want to instead scale by input flux?
            # for reaction in self.model.reactions:
            #     flux = solution.fluxes[reaction.id]
            #     for reaction_name, reaction_flux in state['reaction_fluxes'].items():
            #         if reaction.name == reaction_name:
            #             fluxes[reaction.name] = flux * reaction_flux

        return output_state


# -- fully-spec'd Sed-compliant copasi process --

class SedCobraProcess(Process):
    config_schema = TimeCourseConfig

    def __init__(self, config=None, core=None):
        super().__init__(config, core)
        self.model = load_model(self.config['model']['model_source'])
        self.reactions = self.model.reactions
        self.metabolites = self.model.metabolites
        self.objective = self.model.objective.to_json()['expression']['args'][0]['args'][1]['name']  # TODO -- fix this in cobra
        self.boundary = self.model.boundary

    def initial_state(self):
        solution = self.model.optimize()
        optimized_fluxes = solution.fluxes

        state = {
            'inputs': {
                'reaction_bounds': {}
            },
            'outputs': {
                'fluxes': {}
            }
        }
        for reaction in self.model.reactions:
            state['inputs']['reaction_bounds'][reaction.id] = {
                'lower_bound': reaction.lower_bound,
                'upper_bound': reaction.upper_bound
            }
            state['outputs']['fluxes'][reaction.id] = optimized_fluxes[reaction.id]
        return state

    def inputs(self):
        return {
            'model': 'sbml',
            'reaction_bounds': {
                reaction.id: 'bounds' for reaction in self.reactions
            },
            'objective_reaction': {
                '_type': 'string',
                '_default': self.objective
            },
        }

    def outputs(self):
        return {
            'fluxes': {
                reaction.id: 'float' for reaction in self.reactions
            },
            'objective_value': 'float',
            'reaction_dual_values': {
                reaction.id: 'float' for reaction in self.reactions
            },
            'metabolite_dual_values': {
                metabolite.id: 'float' for metabolite in self.metabolites
            },
            'status': 'string',
        }

    def update(self, inputs, interval):
        # set reaction bounds
        reaction_bounds = inputs['reaction_bounds']
        for reaction_id, bounds in reaction_bounds.items():
            self.model.reactions.get_by_id(reaction_id).bounds = (bounds['lower_bound'], bounds['upper_bound'])

        # set objective
        # TODO -- look into optlang for specifying objective and constraints
        self.model.objective = self.model.reactions.get_by_id(inputs['objective_reaction'])

        # run solver
        solution = self.model.optimize()

        return {
            'fluxes': solution.fluxes.to_dict(),
            'objective_value': solution.objective_value,
            'reaction_dual_values': solution.reduced_costs.to_dict(),
            'metabolite_dual_values': solution.shadow_prices.to_dict(),
            'status': solution.status,
        }


class DynamicFBA(Process):
    """
    Performs dynamic FBA.

    Parameters:
    - model: The metabolic model for the simulation.
    - kinetic_params: Kinetic parameters (Km and Vmax) for each substrate.
    - biomass_reaction: The identifier for the biomass reaction in the model.
    - substrate_update_reactions: A dictionary mapping substrates to their update reactions.
    - biomass_identifier: The identifier for biomass in the current state.

    TODO -- check units
    """

    config_schema = {
        'model_file': 'string',
        'model': 'Any',
        'kinetic_params': 'map[tuple[float,float]]',
        'biomass_reaction': {
            '_type': 'string',
            '_default': 'Biomass_Ecoli_core'
        },
        'substrate_update_reactions': 'map[string]',
        'biomass_identifier': 'string',
        'bounds': 'map[bounds]',
    }

    def __init__(self, config, core):
        super().__init__(config, core)

        if self.config['model_file'] == 'TESTING':
            self.model = MODEL_FOR_TESTING
        elif not 'xml' in self.config['model_file']:
            # use the textbook model if no model file is provided
            self.model = load_model(self.config['model_file'])
        elif isinstance(self.config['model_file'], str):
            self.model = cobra.io.read_sbml_model(self.config['model_file'])
        else:
            # error handling
            raise ValueError('Invalid model file')

        for reaction_id, bounds in self.config['bounds'].items():
            if bounds['lower'] is not None:
                self.model.reactions.get_by_id(reaction_id).lower_bound = bounds['lower']
            if bounds['upper'] is not None:
                self.model.reactions.get_by_id(reaction_id).upper_bound = bounds['upper']

    def inputs(self):
        return {
            'substrates': 'map[positive_float]'
        }

    def outputs(self):
        return {
            'substrates': 'map[positive_float]'
        }

    # TODO -- can we just put the inputs/outputs directly in the function?
    def update(self, state, interval):
        substrates_input = state['substrates']

        for substrate, reaction_id in self.config['substrate_update_reactions'].items():
            Km, Vmax = self.config['kinetic_params'][substrate]
            substrate_concentration = substrates_input[substrate]
            uptake_rate = Vmax * substrate_concentration / (Km + substrate_concentration)
            self.model.reactions.get_by_id(reaction_id).lower_bound = -uptake_rate

        substrate_update = {}

        solution = self.model.optimize()
        if solution.status == 'optimal':
            current_biomass = substrates_input[self.config['biomass_identifier']]
            biomass_growth_rate = solution.fluxes[self.config['biomass_reaction']]
            substrate_update[self.config['biomass_identifier']] = biomass_growth_rate * current_biomass * interval

            for substrate, reaction_id in self.config['substrate_update_reactions'].items():
                flux = solution.fluxes[reaction_id] * current_biomass * interval
                old_concentration = substrates_input[substrate]
                new_concentration = max(old_concentration + flux, 0)  # keep above 0
                substrate_update[substrate] = new_concentration - old_concentration
                # TODO -- assert not negative?
        else:
            # Handle non-optimal solutions if necessary
            # print('Non-optimal solution, skipping update')
            for substrate, reaction_id in self.config['substrate_update_reactions'].items():
                substrate_update[substrate] = 0

        return {
            'substrates': substrate_update,
        }


# register the process
# CORE.register_process('DynamicFBA', DynamicFBA)


# Helper functions to get specs and states
def dfba_config(
        model_file='textbook',
        kinetic_params=None,
        biomass_reaction='Biomass_Ecoli_core',
        substrate_update_reactions=None,
        biomass_identifier='biomass',
        bounds=None
):
    if substrate_update_reactions is None:
        substrate_update_reactions = {
            'glucose': 'EX_glc__D_e',
            'acetate': 'EX_ac_e'}
    if bounds is None:
        bounds = {
            'EX_o2_e': {'lower': -2, 'upper': None},
            'ATPM': {'lower': 1, 'upper': 1}}
    if kinetic_params is None:
        kinetic_params = {
            'glucose': (0.5, 1),
            'acetate': (0.5, 2)}
    return {
        'model_file': model_file,
        'kinetic_params': kinetic_params,
        'biomass_reaction': biomass_reaction,
        'substrate_update_reactions': substrate_update_reactions,
        'biomass_identifier': biomass_identifier,
        'bounds': bounds
    }


def get_single_dfba_spec(mol_ids=None, path=None, i=None, j=None):
    """
    Constructs a configuration dictionary for a dynamic FBA process with optional path indices.

    This function builds a process specification for use with a dynamic FBA system. It allows
    specification of substrate molecule IDs and optionally appends indices to the paths for those substrates.

    Parameters:
        mol_ids (list of str, optional): List of molecule IDs to include in the process. Defaults to
                                         ['glucose', 'acetate', 'biomass'].
        path (list of str, optional): The base path to prepend to each molecule ID. Defaults to ['..', 'fields'].
        i (int, optional): The first index to append to the path for each molecule, if not None.
        j (int, optional): The second index to append to the path for each molecule, if not None.

    Returns:
        dict: A dictionary containing the process type, address, configuration, and paths for inputs
              and outputs based on the specified molecule IDs and indices.
    """
    if path is None:
        path = ['..', 'fields']
    if mol_ids is None:
        mol_ids = ['glucose', 'acetate', 'biomass']

    # Function to build the path with optional indices
    def build_path(mol_id):
        base_path = path + [mol_id]
        if i is not None:
            base_path.append(i)
        if j is not None:
            base_path.append(j)
        return base_path

    return {
        '_type': 'process',
        'address': 'local:DynamicFBA',
        'config': dfba_config(),
        'inputs': {
            'substrates': {mol_id: build_path(mol_id) for mol_id in mol_ids}
        },
        'outputs': {
            'substrates': {mol_id: build_path(mol_id) for mol_id in mol_ids}
        }
    }


def get_spatial_dfba_spec(n_bins=(5, 5), mol_ids=None):
    if mol_ids is None:
        mol_ids = ['glucose', 'acetate', 'biomass']
    dfba_processes_dict = {}
    for i in range(n_bins[0]):
        for j in range(n_bins[1]):
            dfba_processes_dict[f'[{i},{j}]'] = get_single_dfba_spec(mol_ids=mol_ids, path=['..', 'fields'], i=i, j=j)
    return dfba_processes_dict


def get_spatial_dfba_state(
        n_bins=(5, 5),
        mol_ids=None,
        initial_max=None,
):
    if mol_ids is None:
        mol_ids = ['glucose', 'acetate', 'biomass']
    if initial_max is None:
        initial_max = {
            'glucose': 20,
            'acetate': 0,
            'biomass': 0.1
        }
    # initial_fields = {
    #     mol_id: np.random.uniform(low=0, high=initial_max[mol_id], size=n_bins[1], n_bins[0]))
    #     for mol_id in mol_ids}
    initial_fields = {
        mol_id: np.random.uniform(low=0, high=initial_max[mol_id], size=n_bins)
        for mol_id in mol_ids}

    return {
        'fields': {
            '_type': 'map',
            '_value': {
                '_type': 'array',
                '_shape': n_bins,
                '_data': 'positive_float'
            },
            **initial_fields,
        },
        'spatial_dfba': get_spatial_dfba_spec(n_bins=n_bins, mol_ids=mol_ids)
    }


def run_dfba_spatial(
        total_time=60,
        n_bins=(3, 3),  # TODO -- why can't do (5, 10)??
        mol_ids=None,
        CORE=None
):
    if mol_ids is None:
        mol_ids = ['glucose', 'acetate', 'biomass']
    composite_state = get_spatial_dfba_state(
        n_bins=n_bins,
        mol_ids=mol_ids,
    )
    # make the composite
    print('Making the composite...')
    sim = Composite({
        'state': composite_state,
        'emitter': {'mode': 'all'}
    }, core=CORE)

    # save the document
    sim.save(filename='spatial_dfba.json', outdir='out')

    # simulate
    print('Simulating...')
    sim.update({}, total_time)

    # gather results
    dfba_results = sim.gather_results()
    # print(dfba_results)

    print('Plotting results...')
    # plot timeseries
    plot_time_series(
        dfba_results,
        coordinates=[(0, 0), (1, 1), (2, 2)],
        out_dir='out',
        filename='dfba_timeseries.png',
    )

    # make video
    plot_species_distributions_to_gif(
        dfba_results,
        out_dir='out',
        filename='dfba_results.gif',
        title='',
        skip_frames=1
    )



# if __name__ == '__main__':
    # run_dfba_spatial()



import logging
import os
from tempfile import mkdtemp
from typing import *
from logging import Logger

import libsbml
import numpy as np
from process_bigraph.composite import ProcessTypes, Composite
from pymongo import ASCENDING

from bsp.data_model.bigraph import time_course_node_spec
from bsp.compatibility import COMPATIBLE_UTC_SIMULATORS
from bsp.io import normalize_smoldyn_output_path_in_root, get_sbml_species_mapping
from bsp.utils.base_utils import handle_exception, handle_sbml_exception, dynamic_simulator_import

# logging TODO: implement this.
logger: Logger = logging.getLogger("compose.worker.data_generator.log")

# AMICI_ENABLED = True
# COPASI_ENABLED = True
# PYSCES_ENABLED = True
# TELLURIUM_ENABLED = True
# SMOLDYN_ENABLED = True
# READDY_ENABLED = True
# try:
#     from amici import SbmlImporter, import_model_module, Model, runAmiciSimulation
# except ImportError as e:
#     AMICI_ENABLED = False
#     # logger.warning(str(e))
# try:
#     from basico import *
# except ImportError as e:
#     COPASI_ENABLED = False
#     # logger.warning(str(e))
# try:
#     import tellurium as te
# except ImportError as e:
#     TELLURIUM_ENABLED = False
#     # logger.warning(str(e))
# try:
#     from smoldyn import Simulation
#     from smoldyn._smoldyn import MolecState
# except ImportError as e:
#     SMOLDYN_ENABLED = False
#     # logger.warning(str(e))
#
# try:
#     import pysces
# except ImportError as e:
#     PYSCES_ENABLED = False
#     # logger.warning(str(e))

HISTORY_INDEXES = [
    'data.time',
    [('experiment_id', ASCENDING),
     ('data.time', ASCENDING),
     ('_id', ASCENDING)],
]
CONFIGURATION_INDEXES = [
    'experiment_id',
]
SECRETS_PATH = 'secrets.json'


# -- functions related to generating simulation output data (ODE, PDE, stochastic particle, and more) using the process-bigraph engine -- #

def generate_time_course_data(
        input_fp: str,
        start: int,
        end: int,
        steps: int,
        core=None,
        simulators: List[str] = None,
        parameters: Dict[str, Any] = None,
        expected_results_fp: str = None,
        out_dir: str = None
) -> Dict[str, Dict[str, List[float]]]:
    requested_sims = simulators or ["amici", "copasi", "pysces", "tellurium"]
    simulation_spec = {
        simulator: time_course_node_spec(
            input_file=input_fp,
            context=simulator,
            start_time=start,
            end_time=end,
            num_steps=steps
        ) for simulator in requested_sims
    }
    simulation = Composite({'state': simulation_spec, 'emitter': {'mode': 'all'}}, core=core)

    input_filename = input_fp.split("/")[-1].split(".")[0]
    if not out_dir:
        out_dir = mkdtemp()

    if out_dir:
        simulation.save(
            filename=f'{input_filename}-initialization.json',
            outdir=out_dir
        )

    # TODO: is there a better way to do this? (interval of one? Is that symbolic more than anything?)
    if parameters:
        simulation.update(parameters, 1)
    else:
        simulation.run(1)

    if out_dir:
        simulation.save(
            filename=f'{input_filename}-update.json',
            outdir=out_dir
        )

    output_data = {}
    raw_data = simulation.gather_results()[('emitter',)]
    for data in raw_data:
        for data_key, data_value in data.items():
            if data_key.startswith('output_data_store_'):
                simulator = data_key.split('_')[-1]
                output_data[simulator] = data_value

    # return output_data
    import json
    with open(f'{out_dir}/{input_filename}-update.json', 'r') as f:
        state_spec = json.load(f)

    return {'output_data': output_data, 'state': state_spec}


def generate_composition_result_data(
        state_spec: Dict[str, Any],
        duration: int = None,
        core: ProcessTypes = None,
        out_dir: str = None
) -> Dict[str, Union[List[Dict[str, Any]], Dict[str, Any]]]:
    """
    Method used specifically to generate and format the results of composition for use with the REST API.
    """
    simulation = Composite({'state': state_spec, 'emitter': {'mode': 'all'}}, core=core)
    if duration is None:
        duration = 10
    simulation.run(duration)

    results = simulation.gather_results()[('emitter',)]

    import json
    if out_dir is None:
        out_dir = mkdtemp()
    with open(f'{out_dir}/update.json', 'r') as f:
        state_spec = json.load(f)

    return {'results': results, 'state': state_spec}


# -- direct simulator API wrappers -- #

def run_readdy(
        box_size: List[float],
        species_config: List[Dict[str, float]],   # {SPECIES_NAME: DIFFUSION_CONSTANT}  ie: {'E': 10.}
        reactions_config: List[Dict[str, float]],  # {REACTION_SCHEME: REACTION RATE}  ie: {"fwd: E +(0.03) S -> ES": 86.551}
        particles_config: List[Dict[str, Union[List[float], np.ndarray]]],  # {PARTICLE_NAME: INITIAL_POSITIONS_ARRAY}  ie: {'E': np.random.random(size=(n_particles_e, 3)) * edge_length - .5*edge_length}
        dt: float,
        duration: float,
        unit_system_config: Dict[str, str] = None
) -> Dict[str, str]:
    output = {}

    # attempt module install (dynamic)
    readdy = dynamic_simulator_import('readdy')

    # if module error, return error
    if readdy.module is None:
        error = handle_exception("Run Readdy")
        logger.error(error)
        return {'error': error}
    else:
        readdy = readdy.module

    # establish reaction network system
    unit_system = unit_system_config or {"length_unit": "micrometer", "time_unit": "second"}
    system = readdy.ReactionDiffusionSystem(
        box_size=box_size,
        unit_system=unit_system
    )

    # add species via spec
    species_names = []
    for config in species_config:
        species_name = config["name"]
        species_difc = config["diffusion_constant"]
        species_names.append(species_name)
        system.add_species(species_name, diffusion_constant=float(species_difc))

    # add reactions via spec
    for config in reactions_config:
        reaction_scheme = config["scheme"]
        reaction_rate = config["rate"]
        system.reactions.add(reaction_scheme, rate=float(reaction_rate))

    # configure simulation outputs
    simulation = system.simulation(kernel="CPU")
    simulation.output_file = "out.h5"
    simulation.reaction_handler = "UncontrolledApproximation"

    # set initial particle state and configure observations
    for config in particles_config:
        particle_name = config["name"]
        particle_positions = config["initial_positions"]
        if not isinstance(particle_positions, np.ndarray):
            particle_positions = np.array(particle_positions)
        simulation.add_particles(particle_name, particle_positions)

    simulation.observe.number_of_particles(
        stride=1,
        types=list(set(species_names))
    )

    # run simulation for given time parameters
    n_steps = int(float(duration) / dt)
    simulation.run(n_steps=n_steps, timestep=dt)
    output = {"results_file": simulation.output_file}

    return output


# TODO: should we return the actual data from memory, or that reflected in a smoldyn output txt file or both?
def run_smoldyn(model_fp: str, duration: int, dt: float = None) -> Dict[str, Union[str, Dict[str, Union[float, List[float]]]]]:
    """Run the simulation model found at `model_fp` for the duration
        specified therein if output_files are specified in the smoldyn model file and return the aforementioned output file
        or return a dictionary of an array of the `listmols` as well as `molcount` command outputs. NOTE: The model file is currently
        searched for this `output_files` value, and if it exists and not commented out, it will scan the root of the model_fp
        (usually where smoldyn output files are stored, which is the same dir as the model_fp) to retrieve the output file.

            Args:
                model_fp:`str`: path to the smoldyn configuration. Defaults to `None`.
                duration:`float`: duration in seconds to run the simulation for.
                dt:`float`: time step in seconds to run the simulation for. Defaults to None, which uses the built-in simulation dt.

        For the output, we should read the model file and search for "output_files" to start one of the lines.
        If it startswith that, then assume a return of the output txt file, if not: then assume a return from ram.
    """
    # search for output_files in model_fp TODO: optimize this
    # attempt module install (dynamic)
    output_data = {}

    smoldyn = dynamic_simulator_import('smoldyn')
    if smoldyn.module is None:
        error = handle_exception("Run Smoldyn")
        logger.error(error)
        return {'error': error}
    else:
        smoldyn = smoldyn.module

    use_file_output = False
    with open(model_fp, 'r') as f:
        model_content = [line.strip() for line in f.readlines()]
        for content in model_content:
            if content.startswith('output_files'):
                use_file_output = True
        f.close()

    simulation = smoldyn.Simulation.fromFile(model_fp)
    try:
        # case: there is no declaration of output_files in the smoldyn config file, or it is commented out
        if not use_file_output:
            # write molcounts to counts dataset at every timestep (shape=(n_timesteps, 1+n_species <-- one for time)): [timestep, countSpec1, countSpec2, ...]
            simulation.addOutputData('species_counts')
            simulation.addCommand(cmd='molcount species_counts', cmd_type='E')

            # write spatial output to molecules dataset
            simulation.addOutputData('molecules')
            simulation.addCommand(cmd='listmols molecules', cmd_type='E')

            # run simulation for specified time
            step_size = dt or simulation.dt
            simulation.run(duration, step_size, overwrite=True)

            species_count = simulation.count()['species']
            species_names: List[str] = []
            for index in range(species_count):
                species_name = simulation.getSpeciesName(index)
                if 'empty' not in species_name.lower():
                    species_names.append(species_name)

            molecule_output = simulation.getOutputData('molecules')
            counts_output = simulation.getOutputData('species_counts')
            for i, output_array in enumerate(counts_output):
                interval_data = {}
                for j, species_count in enumerate(output_array):
                    interval_data[species_names[j - 1]] = species_count
                counts_output.pop(i)
                counts_output.insert(i, interval_data)

            # return ram data (default dimensions)
            output_data = {'species_counts': counts_output, 'molecules': molecule_output}

        # case: output files are specified, and thus time parameters by which to capture/collect output
        else:
            # run simulation with default time params
            simulation.runSim()

            # change the output filename to a standardized 'modelout.txt' name
            working_dir = os.path.dirname(model_fp)
            results_fp = normalize_smoldyn_output_path_in_root(working_dir)

            # return output file
            output_data = {'results_file': results_fp}
    except:
        error = handle_exception("Run Smoldyn")
        logger.error(error)
        output_data = {'error': error}

    return output_data


def run_sbml_pysces(sbml_fp: str, start: int, dur: int, steps: int) -> Dict[str, Union[List[float], str]]:
    pysces = dynamic_simulator_import('pysces')
    if pysces.module is None:
        error = handle_exception("Run pysces")
        logger.error(error)
        return {'error': error}
    else:
        pysces = pysces.module

    # model compilation
    sbml_filename = sbml_fp.split('/')[-1]
    psc_filename = sbml_filename + '.psc'
    psc_fp = os.path.join(pysces.model_dir, psc_filename)
    # get output with mapping of internal species ids to external (shared) species names
    sbml_species_mapping = get_sbml_species_mapping(sbml_fp)
    obs_names = list(sbml_species_mapping.keys())
    obs_ids = list(sbml_species_mapping.values())
    # run the simulation with specified time params and get the data
    try:
        # NOTE: the below model load works only in pysces 1.2.2 which is not available on conda via mac. TODO: fix this.
        model = pysces.loadSBML(sbmlfile=sbml_fp, pscfile=psc_fp)
        model.sim_time = np.linspace(start, dur, steps + 1)
        model.Simulate(1)  # specify userinit=1 to directly use model.sim_time (t) rather than the default
        return {
            name: model.data_sim.getSimData(obs_id)[:, 1].tolist()
            for name, obs_id in sbml_species_mapping.items()
        }
    except:
        error_message = handle_sbml_exception()
        logger.error(error_message)
        return {"error": error_message}


def run_sbml_tellurium(sbml_fp: str, start: int, dur: int, steps: int) -> Dict[str, Union[List[float], str]]:
    te = dynamic_simulator_import('tellurium')
    if te.module is None:
        error = handle_exception("Run tellurium")
        logger.error(error)
        return {'error': error}
    else:
        te = te.module

    result = None
    try:
        simulator = te.loadSBMLModel(sbml_fp)
        if start > 0:
            simulator.simulate(0, start)
        result = simulator.simulate(start, dur, steps + 1)
        species_mapping = get_sbml_species_mapping(sbml_fp)
        if result is not None:
            outputs = {}
            for colname in result.colnames:
                if 'time' not in colname:
                    for spec_name, spec_id in species_mapping.items():
                        if colname.replace("[", "").replace("]", "") == spec_id:
                            data = result[colname]
                            outputs[spec_name] = data.tolist()
            return outputs
        else:
            raise Exception('Tellurium: Could not generate results.')
    except:
        error_message = handle_sbml_exception()
        logger.error(error_message)
        return {"error": error_message}


def run_sbml_copasi(sbml_fp: str, start: int, dur: int, steps: int) -> Dict[str, Union[List[float], str]]:
    basico = dynamic_simulator_import('basico')
    if basico.module is None:
        error = handle_exception("Run basico")
        logger.error(error)
        return {'error': error}
    else:
        basico = basico.module

    try:
        t = np.linspace(start, dur, steps + 1)
        model = basico.load_model(sbml_fp)
        specs = basico.get_species(model=model).index.tolist()
        for spec in specs:
            if spec == "EmptySet" or "EmptySet" in spec:
                specs.remove(spec)
        tc = basico.run_time_course(model=model, update_model=True, values=t)
        data = {spec: tc[spec].values.tolist() for spec in specs}
        return data
    except:
        error_message = handle_sbml_exception()
        logger.error(error_message)
        return {"error": error_message}


def run_sbml_amici(sbml_fp: str, start: int, dur: int, steps: int) -> Dict[str, Union[List[float], str]]:
    amici = dynamic_simulator_import('amici')
    if amici.module is None:
        error = handle_exception("Run amici")
        logger.error(error)
        return {'error': error}
    else:
        amici = amici.module

    try:
        sbml_reader = libsbml.SBMLReader()
        sbml_doc = sbml_reader.readSBML(sbml_fp)
        sbml_model_object = sbml_doc.getModel()
        sbml_importer = amici.SbmlImporter(sbml_fp)
        model_id = sbml_fp.split('/')[-1].replace('.xml', '')
        model_output_dir = mkdtemp()
        sbml_importer.sbml2amici(
            model_id,
            model_output_dir,
            verbose=logging.INFO,
            observables=None,
            sigmas=None,
            constant_parameters=None
        )
        # model_output_dir = model_id  # mkdtemp()
        model_module = amici.import_model_module(model_id, model_output_dir)
        amici_model_object = model_module.getModel()
        floating_species_list = list(amici_model_object.getStateIds())
        floating_species_initial = list(amici_model_object.getInitialStates())
        sbml_species_ids = [spec.getName() for spec in sbml_model_object.getListOfSpecies()]
        t = np.linspace(start, dur, steps + 1)
        amici_model_object.setTimepoints(t)
        initial_state = dict(zip(floating_species_list, floating_species_initial))
        set_values = []
        for species_id, value in initial_state.items():
            set_values.append(value)
        amici_model_object.setInitialStates(set_values)
        sbml_species_mapping = get_sbml_species_mapping(sbml_fp)
        method = amici_model_object.getSolver()
        result_data = amici.runAmiciSimulation(solver=method, model=amici_model_object)
        results = {}
        floating_results = dict(zip(
            sbml_species_ids,
            list(map(
                lambda x: result_data.by_id(x),
                floating_species_list
            ))
        ))
        results = floating_results
        return {
            key: val.tolist() if isinstance(val, np.ndarray) else val
            for key, val in results.items()
        }
    except:
        error_message = handle_sbml_exception()
        logger.error(error_message)
        return {"error": error_message}


# TODO: add vcell and masspy here
SBML_EXECUTORS = dict(zip(
    [data[0] for data in COMPATIBLE_UTC_SIMULATORS],
    [run_sbml_amici, run_sbml_copasi, run_sbml_pysces, run_sbml_tellurium]
))


# -- formatted observables data -- #

def generate_sbml_outputs(sbml_fp: str, start: int, dur: int, steps: int, simulators: list[str] = None, truth: str = None) -> dict:
    # TODO: add VCELL and pysces here
    output = {}
    sbml_species_ids = list(get_sbml_species_mapping(sbml_fp).keys())
    simulators = simulators or ['amici', 'copasi', 'tellurium', 'pysces']
    all_output_ids = []
    for simulator in simulators:
        results = {}
        simulator = simulator.lower()
        simulation_executor = SBML_EXECUTORS[simulator]
        sim_result = simulation_executor(sbml_fp=sbml_fp, start=start, dur=dur, steps=steps)

        # case: simulation execution was successful
        if "error" not in sim_result.keys():
            # add to all shared names
            all_output_ids.append(list(sim_result.keys()))

            # iterate over sbml_species_ids to index output data
            for species_id in sbml_species_ids:
                if species_id in sim_result.keys():
                    output_vals = sim_result[species_id]
                    if isinstance(output_vals, np.ndarray):
                        output_vals = output_vals.tolist()
                    results[species_id] = output_vals
        else:
            # case: simulation had an error
            results = sim_result

        # set the simulator output
        output[simulator] = results

    # get the commonly shared output ids
    final_output = {}
    shared_output_ids = min(all_output_ids)
    for simulator_name in output.keys():
        sim_data = {}
        for spec_id in output[simulator_name].keys():
            if spec_id in shared_output_ids:
                sim_data[spec_id] = output[simulator_name][spec_id]
            elif spec_id == "error":
                sim_data["error"] = output[simulator_name][spec_id]

        final_output[simulator_name] = sim_data

    return final_output


# -- process-bigraph implementations -- #






"""The output data returned by that which is required by simularium (executiontime, listmols),
    when written and read into the same file for a given global time is as follows:

    [identity, state, x, y, z, serial number], where:

        identity = species identity for molecule
        state = state of the given molecule
        x, y, z = values for the relative coordinates
        serial_number = monotonically decreasing timestamp for the given species_id

        At each global timestep (`executiontime`), a new 'cast of characters' are introduced that may resemble the
            cast of characters at the first timestep, but are in fact different and thus all the molecules provided
            from the `listmols` command will in fact all be unique.


    I propose the following consideration at each `update` step:

    The `smoldyn.Simulation().connect()` method can be used to (for example) tell the
    simulation about external environmental parameters that change over the course of
    the simulation. For example:

        a = None
        avals = []

        def new_difc(t, args):
            global a, avals
            x, y = args
            avals.append((t, a.difc['soln']))
            return x * math.sin(t) + y

        def update_difc(val):
            global a
            a.difc = val

        def test_connect():
            global a, avals
            sim = smoldyn.Simulation(low=(.......
            a = sim.addSpecies('a', color=black, difc=0.1)

            # specify either function as target:
            sim.connect(new_dif, update_difc, step=10, args=[1,1])

            # ...or species as target:
            sim.connect(func = new_dif, target = 'a.difc', step=10, args=[0, 1])

            sim.run(....


"""
import math
import os
from dataclasses import dataclass
from typing import *
from uuid import uuid4

import numpy as np
from process_bigraph import Process, Composite, pf, pp

from bsp.data_model.base import BaseClass
from bsp.data_model.sed import SedModel

try:
    import smoldyn as sm
    from smoldyn._smoldyn import MolecState, Simulation
except:
    raise ImportError(
        '\nPLEASE NOTE: Smoldyn is not correctly installed on your system which prevents you from ' 
        'using the SmoldynProcess. Please refer to the README for further information '
        'on installing Smoldyn.'
    )


class SmoldynProcess(Process):
    config_schema = {
        'model': 'SedModelConfig',
        'animate': {
            '_type': 'boolean',
            '_default': False
        }
    }

    def __init__(self, config: Dict[str, Any] = None, core=None):
        """A new instance of `SmoldynProcess` based on the `config` that is passed. The schema for the config to be passed in
            this object's constructor is as follows:

            config_schema = {
                'model_filepath': 'string',  <-- analogous to python `str`
                'animate': 'bool'  <-- of type `bigraph_schema.base_types.bool`

            # TODO: It would be nice to have classes associated with this.
        """
        super().__init__(config, core)

        # specify the model fp for clarity
        # self.model_filepath = self.config.get('model_filepath')
        self.model_filepath = self.config.get('model').get('model_source')

        # enforce model filepath passing
        if not self.model_filepath:
            raise ValueError(
                '''
                    The Process configuration requires a Smoldyn model filepath to be passed.
                    Please specify a 'model_filepath' in your instance configuration.
                '''
            )

        # initialize the simulator from a Smoldyn MinE.txt file.
        self.simulation: sm.Simulation = sm.Simulation.fromFile(self.model_filepath)

        # get a list of the simulation species
        species_count = self.simulation.count()['species']
        self.species_names: List[str] = []
        for index in range(species_count):
            species_name = self.simulation.getSpeciesName(index)
            if 'empty' not in species_name.lower():
                self.species_names.append(species_name)
        # sort for logistical mapping to species names (i.e: ['a', 'b', c'] == ['0', '1', '2']
        self.species_names.sort()

        # make species counts of molecules dataset for output
        self.simulation.addOutputData('species_counts')
        # write molcounts to counts dataset at every timestep (shape=(n_timesteps, 1+n_species <-- one for time)): [timestep, countSpec1, countSpec2, ...]
        self.simulation.addCommand(cmd='molcount species_counts', cmd_type='E')

        # make molecules dataset (molecule information) for output
        self.simulation.addOutputData('molecules')
        # write coords to dataset at every timestep (shape=(n_output_molecules, 7)): seven being [timestep, smol_id(species), mol_state, x, y, z, mol_serial_num]
        self.simulation.addCommand(cmd='listmols2 molecules', cmd_type='E')

        # initialize the molecule ids based on the species names. We need this value to properly emit the schema, which expects a single value from this to be a str(int)
        # the format for molecule_ids is expected to be: 'speciesId_moleculeNumber'
        self.molecule_ids: List[str] = [str(uuid4()) for n in list(range(len(self.species_names)))]

        # get the simulation boundaries, which in the case of Smoldyn denote the physical boundaries
        # TODO: add a verification method to ensure that the boundaries do not change on the next step...
        self.boundaries: Dict[str, List[float]] = dict(zip(['low', 'high'], self.simulation.getBoundaries()))

        # set graphics (defaults to False)
        if self.config['animate']:
            self.simulation.addGraphics('opengl_better')

        # create a re-usable counts and molecules type to be used by both inputs and outputs
        self.counts_type = {
            species_name: 'integer'
            for species_name in self.species_names
        }

        self.port_schema = {
            'species_counts': {
                species_name: 'integer'
                for species_name in self.species_names
            },
            'molecules': 'tree[float]',  # self.molecules_type
            'geometry': 'GeometryType',
            'forces': 'list',
        }

        self._specs = [None for _ in self.species_names]
        self._vals = dict(zip(self.species_names, [[] for _ in self.species_names]))

    def initial_state(self) -> Dict[str, Union[int, Dict]]:
        """Set the initial parameter state of the simulation. This method should return an implementation of
            that which is returned by `self.schema()`.


        NOTE: Due to the nature of this model,
            Smoldyn assigns a random uniform distribution of integers as the initial coordinate (x, y, z)
            values for the simulation. As such, the `set_uniform` method will uniformly distribute
            the molecules according to a `highpos`[x,y] and `lowpos`[x,y] where high and low pos are
            the higher and lower bounds of the molecule spatial distribution.

            NOTE: This method should provide an implementation of the structure denoted in `self.schema`.
        """
        # get the initial species counts
        initial_species_counts = {
            spec_name: self.simulation.getMoleculeCount(spec_name, MolecState.all)
            for spec_name in self.species_names
        }

        return {
            'geometry': {
                'faces': [],
                'vertices': []
            },
            'forces': [],
            'species_counts': initial_species_counts,
            'molecules': {
                mol_id: {
                    'coordinates': [0.0, 0.0, 0.0],
                    'species_id': self.species_names[i],
                    'state': 0
                }
                for i, mol_id in enumerate(self.molecule_ids)
            }
        }

    def inputs(self):
        species_counts_schema = {
            species_name: 'integer'
            for species_name in self.species_names
        }
        return {
            'species_counts': species_counts_schema,
            'particles': 'ParticleType',
            'geometry': 'GeometryType',
            'net_forces': 'MechanicalForcesType'
        }

    def outputs(self):
        species_counts_schema = {
            species_name: 'integer'
            for species_name in self.species_names
        }
        return {
            'species_counts': species_counts_schema,
            'particles': 'ParticleType',
            # 'geometry': 'GeometryType',
            # 'net_forces': 'MechanicalForcesType'
        }

    def update(self, state: Dict, interval: int) -> Dict:
        """TODO: We must account for the mol_ids that are generated in the output based on the interval run,
                i.e: Shorter intervals will yield both less output molecules and less unique molecule ids.
        """
        # take in geometry from state
        vertices_k = np.array(state['geometry']['vertices'])

        # get upper and lower boundary coords from vertices: iterate over each axis and get max of each axis
        bounds_low, bounds_high = get_kth_boundaries(vertices_k)

        # reset the molecules, distribute the mols according to dynamic bounds
        for name in self.species_names:
            set_uniform(
                simulation=self.simulation,
                species_name=name,
                boundary_high=max(bounds_high),
                boundary_low=min(bounds_low),
                count=state['species_counts'][name],
                kill_mol=False
            )

            # TODO: extract the difc from the model somehow!
            # d0 = self.simulation.getSpecies(name).difc
            d0 = 0.3  # placeholder difc
            forces_k = np.array(state['net_forces'])
            alpha = 0.3  # TODO: make this not arbitrary
            beta = 1.0

            # TODO: we need to take in noticeable vertices from the membrane force prescription and infer beta for given particle based on its coords at iteration k
            # derive beta based on the given particle's distance at iteration k from the membrane (bounds)

            self.simulation.connect(
                func=compute_kth_difc,
                target=f'{name}.difc',
                step=interval,
                args=[d0, forces_k, alpha, beta],
            )

        # run the simulation for a given interval
        self.simulation.run(
            stop=interval,
            dt=self.simulation.dt
        )

        # get the counts data, clear the buffer
        counts_data = self.simulation.getOutputData('species_counts')

        # get the final counts for the update
        final_count = counts_data[-1]

        # remove the timestep from the list
        final_count.pop(0)

        # get the data based on the commands added in the constructor, clear the buffer
        molecules_data = self.simulation.getOutputData('molecules')

        # create an empty simulation state mirroring that which is specified in the schema
        simulation_state = {
            'species_counts': {},
            'particles': {}
        }

        # get and populate the species counts
        for index, name in enumerate(self.species_names):
            input_counts = state['species_counts'][name]
            simulation_state['species_counts'][name] = int(final_count[index]) - input_counts

        # clear the list of known molecule ids and update the list of known molecule ids (convert to an intstring)
        self.molecule_ids.clear()
        for molecule in molecules_data:
            self.molecule_ids.append(str(uuid4()))

        # get and populate the output molecules
        for index, mol_id in enumerate(self.molecule_ids):
            single_molecule_data = molecules_data[index]
            single_molecule_species_index = int(single_molecule_data[1]) - 1

            simulation_state['particles'][mol_id] = {
                'coordinates': single_molecule_data[3:6],
                'species_id': self.species_names[single_molecule_species_index],
                'state': str(int(single_molecule_data[2]))
            }

        # TODO -- post processing to get effective rates

        return simulation_state


# TODO: finish this
@dataclass
class Boundary(BaseClass):
    low: tuple[list[float], list[float]]
    high: tuple[list[float], list[float]]


def get_axis_vertices(vertices, axis: int, func):
    return func([vertex[axis] for vertex in vertices])


def get_bounds(vertices, axis: int) -> tuple[float, float]:
    ax_min = get_axis_vertices(vertices, axis, min)
    ax_max = get_axis_vertices(vertices, axis, max)
    return ax_min, ax_max


def get_kth_boundaries(vertices) -> tuple[list[float], list[float]]:
    """

    """
    x_min, x_max = get_bounds(vertices, 0)
    y_min, y_max = get_bounds(vertices, 1)
    z_min, z_max = get_bounds(vertices, 2)

    return [x_min, y_min, z_min], [x_max, y_max, z_max]


def compute_kth_difc(t, args):
    """
    Called by `smoldyn.Simulation` during the kth iteration (t) of `SmoldynProcess.update()`.

    :param t: time step of kth iteration.
    :param args: arguments required to compute an updated difc for each particle at iteration k in `SmoldynProcess.update()`. The
        arguments are:
        D0: difc for species s at iteration k.
        forces: the force vector matrix from the membrane at iteration k.
        alpha: membrane vertex force sensitivity constant based on membrane construction and make-up.
        beta: a force sensitivity constant based on particle location from membrane.
    """
    # TODO: make alpha and beta more formal/expandable terms
    # alpha = 0.5, beta = 1.0

    D0, forces, distance, alpha, beta = args
    if not isinstance(forces, np.ndarray):
        forces = np.array(forces)

    force_magnitude = np.linalg.norm(forces)
    alpha_scaled = alpha * np.exp(-beta * distance)

    return D0 * np.exp(-alpha_scaled * force_magnitude)


def set_uniform(
        simulation: Simulation,
        species_name: str,
        count: int,
        boundary_high: float,
        boundary_low: float,
        kill_mol: bool = True
) -> None:
    """Add a distribution of molecules to the solution in
        the simulation memory given a higher and lower bound x,y coordinate. Smoldyn assumes
        a global boundary versus individual species boundaries. Kills the molecule before dist if true.
        TODO: If pymunk expands the species compartment, account for
              expanding `highpos` and `lowpos`. This method should be used within the body/logic of
              the `update` class method.
    """
    # kill the mol, effectively resetting it
    if kill_mol:
        simulation.runCommand(f'killmol {species_name}')

    # TODO: eventually allow for an expanding boundary ie in the configuration parameters (pymunk?), which is defies the methodology of smoldyn

    # redistribute the molecule according to the bounds
    return simulation.addSolutionMolecules(
        species=species_name,
        number=count,
        highpos=boundary_high,
        lowpos=boundary_low
    )




# -- SMOLDYN IO PROCESS FOR SIMULARIUM -- #

class SmoldynIOProcess(Process):
    """
    Parameters:
        model: model: {model_source: INPUT FILEPATH, ...}
        duration: simulation duration (int)
        output_dest: tmp output dir (str)
    """
    config_schema = {
        'model': SedModel,
        # 'duration': 'integer',  # duration should instead be inferred from Composite
        'output_dest': 'string',
        'animate': {
            '_type': 'boolean',
            '_default': False
        }
    }

    def __init__(self, config: dict = None, core=None):
        super().__init__(config, core)

        # get params
        input_filepath = self.config.get('model').get('model_source')
        input_filename = input_filepath.split('/')[-1]
        output_dest = self.config['output_dest']
        self.model_filepath = os.path.join(output_dest, input_filename)
        # self.duration = self.config['duration']

        # write the uploaded file to save dest(temp)
        with open(input_filepath, 'r') as source_file:
            with open(self.model_filepath, 'w') as model_file:
                # Read from the source file and write to the destination
                model_file.write(source_file.read())

        # get the appropriate output filepath
        self.output_filepath = self.handle_output_commands()

        # enforce model filepath passing
        if not self.model_filepath:
            raise ValueError(
                '''
                    The Process configuration requires a Smoldyn model filepath to be passed.
                    Please specify a 'model_filepath' in your instance configuration.
                '''
            )

        # initialize the simulator from a Smoldyn MinE.txt file.
        self.simulation: sm.Simulation = sm.Simulation.fromFile(self.model_filepath)

        # get a list of the simulation species
        species_count = self.simulation.count()['species']
        self.species_names: List[str] = []
        for index in range(species_count):
            species_name = self.simulation.getSpeciesName(index)
            if 'empty' not in species_name.lower():
                self.species_names.append(species_name)
        # sort for logistical mapping to species names (i.e: ['a', 'b', c'] == ['0', '1', '2']
        self.species_names.sort()
        self.port_schema = {'output_filepath': 'string'}
        self.boundaries: Dict[str, List[float]] = dict(zip(['low', 'high'], self.simulation.getBoundaries()))
        self.interval = 0

    def set_uniform(
            self,
            species_name: str,
            count: int,
            kill_mol: bool = True
    ) -> None:
        """Add a distribution of molecules to the solution in
            the simulation memory given a higher and lower bound x,y coordinate. Smoldyn assumes
            a global boundary versus individual species boundaries. Kills the molecule before dist if true.

            TODO: If pymunk expands the species compartment, account for
                  expanding `highpos` and `lowpos`. This method should be used within the body/logic of
                  the `update` class method.

            Args:
                species_name:`str`: name of the given molecule.
                count:`int`: number of molecules of the given `species_name` to add.
                kill_mol:`bool`: kills the molecule based on the `name` argument, which effectively
                    removes the molecule from simulation memory.
        """
        # kill the mol, effectively resetting it
        if kill_mol:
            self.simulation.runCommand(f'killmol {species_name}')

        # TODO: eventually allow for an expanding boundary ie in the configuration parameters (pymunk?), which is defies the methodology of smoldyn

        # redistribute the molecule according to the bounds
        self.simulation.addSolutionMolecules(
            species=species_name,
            number=count,
            highpos=self.boundaries['high'],
            lowpos=self.boundaries['low']
        )

    def initial_state(self) -> Dict[str, Union[int, Dict]]:
        """Set the initial parameter state of the simulation. This method should return an implementation of
            that which is returned by `self.schema()`.


        NOTE: Due to the nature of this model,
            Smoldyn assigns a random uniform distribution of integers as the initial coordinate (x, y, z)
            values for the simulation. As such, the `set_uniform` method will uniformly distribute
            the molecules according to a `highpos`[x,y] and `lowpos`[x,y] where high and low pos are
            the higher and lower bounds of the molecule spatial distribution.

            NOTE: This method should provide an implementation of the structure denoted in `self.schema`.
        """
        return {}

    def inputs(self):
        return {}

    def outputs(self):
        return self.port_schema

    def update(self, inputs: Dict, interval: int) -> Dict:
        """Callback method to be evoked at each Process interval. We want to get the
            last of each dataset type as that is the relevant data in regard to the Process timescale scope.

            Args:
                inputs:`Dict`: current state of the Smoldyn simulation, expressed as a `Dict` whose
                    schema matches that which is returned by the `self.schema()` API method.
                interval:`int`: Analogous to Smoldyn's `time_stop`, this is the
                    timestep interval at which to provide the update as the output of this method.
                    NOTE: This update is iteratively called with the `Process` API.

            Returns:
                `Dict`: New state according to the update at interval


            TODO: We must account for the mol_ids that are generated in the output based on the interval run,
                i.e: Shorter intervals will yield both less output molecules and less unique molecule ids.
        """
        # get current counts state from simulator instance
        species_counts = {
            spec_name: self.simulation.getMoleculeCount(spec_name, MolecState.all)
            for spec_name in self.species_names
        }

        # reset the molecules, distribute the mols according to self.boundarieså
        for name in self.species_names:
            # TODO: find a better MolecState to use
            current_species_count = self.simulation.getMoleculeCount(name, MolecState.all)
            self.set_uniform(
                species_name=name,
                count=current_species_count,
                kill_mol=True
            )

        # run the simulation for a given interval
        # self.simulation.run(
        #     stop=interval + 1,
        #     dt=self.simulation.dt
        # )
        self.simulation.runSim()

        # rewrite the modelout to a unique name
        outfile = self.output_filepath.split('/')[-1]
        output_filename = str(self.interval) + "_" + outfile
        output_filepath = self.output_filepath.replace(outfile, output_filename)
        self.interval += 1

        # return the modelout file
        return {'output_filepath': output_filepath}

    def read_smoldyn_simulation_configuration(self, filename) -> list[str]:
        ''' Read a configuration for a Smoldyn simulation

        Args:
            filename (:obj:`str`): path to model file

        Returns:
            :obj:`list` of :obj:`str`: simulation configuration
        '''
        with open(filename, 'r') as file:
            return [line.strip('\n') for line in file]

    def write_smoldyn_simulation_configuration(self, configuration, filename) -> None:
        ''' Write a configuration for Smoldyn simulation to a file

        Args:
            configuration
            filename (:obj:`str`): path to save configuration
        '''
        with open(filename, 'w') as file:
            for line in configuration:
                file.write(line)
                file.write('\n')

    def handle_output_commands(self) -> str:
        model_fp = self.model_filepath
        duration = 1
        config = self.read_smoldyn_simulation_configuration(model_fp)
        has_output_commands = any([v.startswith('output') for v in config])
        if not has_output_commands:
            cmd_i = 0
            for i, v in enumerate(config):
                if v == 'end_file':
                    cmd_i += i - 1
                stop_key = 'time_stop'

                if f'define {stop_key.upper()}' in v:
                    new_v = f'define TIME_STOP   {duration}'
                    config.remove(config[i])
                    config.insert(i, new_v)
                elif v.startswith(stop_key):
                    new_v = f'time_stop TIME_STOP'
                    config.remove(config[i])
                    config.insert(i, new_v)
            cmds = ["output_files modelout.txt",
                    "cmd i 0 TIME_STOP 1 executiontime modelout.txt",
                    "cmd i 0 TIME_STOP 1 listmols modelout.txt"]
            current = cmd_i
            for cmd in cmds:
                config.insert(current, cmd)
                current += 1

        self.write_smoldyn_simulation_configuration(config, model_fp)

        out_filepath = model_fp.replace(model_fp.split('/')[-1].split('.')[0], 'modelout')
        return out_filepath

    def _new_difc(self, t, args):
        minD_count, minE_count = args
        # TODO: extract real volume
        volume = 1.0

        # Calculate concentrations from counts
        minD_conc = minD_count / volume
        minE_conc = minE_count / volume

        # for example: diffusion coefficient might decrease with high MinD-MinE complex formation
        minDE_complex = minD_conc * minE_conc  # Simplified interaction term
        base_diffusion = 2.5  # Baseline diffusion coefficient

        # perturb the parameter of complex formation effect slightly
        _alpha = 0.1
        delta = _alpha - (_alpha ** 10)
        alpha = _alpha - delta

        new_difc = base_diffusion * (1 - alpha * minDE_complex)

        # Ensure the diffusion coefficient remains within a reasonable range
        new_difc = max(min(new_difc, base_diffusion), 0.01)  # Adjust bounds as needed

        return new_difc


def test_process(core):
    """Test the smoldyn process using the crowding model."""

    # this is the instance for the composite process to run
    print("RUNNING")
    core.process_registry.register('biosimulators_processes.processes.smoldyn_process.SmoldynProcess', SmoldynProcess)
    instance = {
        'smoldyn': {
            '_type': 'process',
            'address': 'local:!biosimulators_processes.processes.smoldyn_process.SmoldynProcess',
            'config': {
                'model_filepath': 'biosimulators_processes/model_files/minE_model.txt',
                'animate': False},
            'inputs': {
                'species_counts': ['species_counts_store'],
                'molecules': ['molecules_store']},
            'outputs': {
                'species_counts': ['species_counts_store'],
                'molecules': ['molecules_store']}
        },
        'emitter': {
            '_type': 'step',
            'address': 'local:ram-emitter',
            'config': {
                'emit': {
                    'species_counts': 'tree[string]',
                    'molecules': 'tree[string]'}
            },
            'inputs': {
                'species_counts': ['species_counts_store'],
                'molecules': ['molecules_store']}
        }
    }

    total_time = 2

    # make the composite
    # workflow = Composite({
    #     'state': instance
    # })

    # run
    # workflow.run(total_time)

    # gather results
    # results = workflow.gather_results()
    # pp(f'RESULTS: {pf(results)}')



from math import pi
from typing import *

import numpy as np
import rustworkx as rx
import networkx as nx
from qiskit_nature.second_q.hamiltonians.lattices import (
    BoundaryCondition,
    HyperCubicLattice,
    Lattice,
    LatticeDrawStyle,
    LineLattice,
    SquareLattice,
    TriangularLattice,
)
from qiskit_nature.second_q.hamiltonians import FermiHubbardModel
from qiskit_nature.second_q.problems import LatticeModelProblem
from qiskit_algorithms import NumPyMinimumEigensolver
from qiskit_nature.second_q.algorithms import GroundStateEigensolver
from qiskit_nature.second_q.mappers import JordanWignerMapper
from qiskit.quantum_info import Pauli, SparsePauliOp
from qiskit.primitives import Sampler
from qiskit.quantum_info import Pauli
from qiskit.result import QuasiDistribution
from qiskit_algorithms import QAOA
from qiskit_algorithms.optimizers import COBYLA
from qiskit_algorithms.utils import algorithm_globals

from process_bigraph import Composite, Process
# TODO: import more qiskit nature for sbml processes here
from biosimulators_processes import CORE


class QiskitProcess(Process):
    config_schema = {
        'num_qbits': 'int',
        'duration': 'int'
    }

    def __init__(self, config=None, core=CORE):
        super().__init__(config, core)

    def initial_state(self):
        return {}

    def inputs(self):
        return {}

    def outputs(self):
        return {}

    def update(self, state, interval):
        return {}


class QAOAProcess(QiskitProcess):
    config_schema = {
        'bigraph_instance': 'tree[any]'}

    def __init__(self, config=None, core=CORE):
        # TODO: Finish this based on https://qiskit-community.github.io/qiskit-algorithms/tutorials/05_qaoa.html
        super().__init__(config, core)
        self.num_nodes = len(list(self.config['bigraph_instance'].keys()))

        # TODO: Enable dynamic setting of these weights with np.stack
        weights = []
        for i, n in enumerate(list(range(self.num_nodes))):
            adj = [0.0, 1.0, 1.0, 0.0]  # TODO: Finish this below
            weights.append(adj)

        # self.w = np.stack(weights)
        self.w = np.array([
            [0.0, 1.0, 1.0, 0.0],
            [1.0, 0.0, 1.0, 1.0],
            [1.0, 1.0, 0.0, 1.0],
            [0.0, 1.0, 1.0, 0.0]])

        self.G = nx.from_numpy_array(self.w)
        self.sampler = Sampler()

        print('Drawing graph: ')
        self._draw_graph()

    def initial_state(self):
        return {}

    def inputs(self):
        return {'weights': 'list[float]', 'objective_result': 'int'}

    def outputs(self):
        return {'weights': 'list[float]', 'objective_result': 'int'}

    def update(self, state, interval):
        qubit_op, offset = self._get_operator(self.w)
        algorithm_globals.random_seed = 10598

        optimizer = COBYLA()
        qaoa = QAOA(self.sampler, optimizer, reps=2)

        result = qaoa.compute_minimum_eigenvalue(qubit_op)

        x = self._sample_most_likely(result.eigenstate)
        objective_result = self._objective_value(x, self.w)
        # TODO: somehow update the weights assigned to self.w here.
        return {'weights': self.w, 'objective_result': objective_result}

    @classmethod
    def _sample_most_likely(cls, state_vector):
        """Compute the most likely binary string from state vector.
        Args:
            state_vector: State vector or quasi-distribution.

        Returns:
            Binary string as an array of ints.
        """
        if isinstance(state_vector, QuasiDistribution):
            values = list(state_vector.values())
        else:
            values = state_vector
        n = int(np.log2(len(values)))
        k = np.argmax(np.abs(values))
        x = cls._bitfield(k, n)
        x.reverse()
        return np.asarray(x)

    @classmethod
    def _objective_value(cls, x, w):
        """Compute the value of a cut.
        Args:
            x: Binary string as numpy array.
            w: Adjacency matrix.
        Returns:
            Value of the cut.
        """
        X = np.outer(x, (1 - x))
        w_01 = np.where(w != 0, 1, 0)
        return np.sum(w_01 * X)

    @classmethod
    def _bitfield(cls, n, L):
        result = np.binary_repr(n, L)
        return [int(digit) for digit in result]

    @classmethod
    def _get_operator(cls, weight_matrix) -> Tuple[SparsePauliOp, int]:
        r"""Generate Hamiltonian for the graph partitioning
        Notes:
            Goals:
                1 Separate the vertices into two set of the same size.
                2 Make sure the number of edges between the two set is minimized.
            Hamiltonian:
                H = H_A + H_B
                H_A = sum\_{(i,j)\in E}{(1-ZiZj)/2}
                H_B = (sum_{i}{Zi})^2 = sum_{i}{Zi^2}+sum_{i!=j}{ZiZj}
                H_A is for achieving goal 2 and H_B is for achieving goal 1.
        Args:
            weight_matrix: Adjacency matrix.
        Returns:
            Operator for the Hamiltonian
            A constant shift for the obj function.
        """
        num_nodes = len(weight_matrix)
        pauli_list = []
        coeffs = []
        shift = 0

        for i in range(num_nodes):
            for j in range(i):
                if weight_matrix[i, j] != 0:
                    x_p = np.zeros(num_nodes, dtype=bool)
                    z_p = np.zeros(num_nodes, dtype=bool)
                    z_p[i] = True
                    z_p[j] = True
                    pauli_list.append(Pauli((z_p, x_p)))
                    coeffs.append(-0.5)
                    shift += 0.5

        for i in range(num_nodes):
            for j in range(num_nodes):
                if i != j:
                    x_p = np.zeros(num_nodes, dtype=bool)
                    z_p = np.zeros(num_nodes, dtype=bool)
                    z_p[i] = True
                    z_p[j] = True
                    pauli_list.append(Pauli((z_p, x_p)))
                    coeffs.append(1.0)
                else:
                    shift += 1

        return SparsePauliOp(pauli_list, coeffs=coeffs), shift

    def _draw_graph(self):
        layout = nx.random_layout(self.G, seed=10)
        colors = ["r", "g", "b", "y"]
        nx.draw(self.G, layout, node_color=colors)
        labels = nx.get_edge_attributes(self.G, "weight")
        return nx.draw_networkx_edge_labels(self.G, pos=layout, edge_labels=labels)


class QuantumAutoencoderProcess(QiskitProcess):
    config_schema = {
        'num_qbits': 'int',
        'duration': 'int',
        'global_random_seed': {
            '_default': 42,
            '_type': 'string'
        }}

    def __init__(self, config=None, core=CORE):
        super().__init__(config, core)
        """
            TODO: See https://qiskit-community.github.io/qiskit-machine-learning/tutorials/12_quantum_autoencoder.html
        """


class LatticeGroundStateSolverProcess(QiskitProcess):
    config_schema = {
        'num_nodes': 'int',
        'interaction_parameters': {
            '_type': 'tree[float]',
            '_default': {'t': -1.0, 'u': 5.0}},
        'onsite_potential': {  # <--alias for v
            '_default': 0.0,
            '_type': 'float'}}

    def __init__(self, config=None, core=CORE):
        super().__init__(config, core)

        num_nodes = self.config['num_nodes']
        boundary_condition = BoundaryCondition.OPEN
        line_lattice = LineLattice(num_nodes=num_nodes, boundary_condition=boundary_condition)
        interaction_parameters = self.config['interaction_parameters']
        t = interaction_parameters['t']
        u = interaction_parameters['u']
        v = self.config['onsite_potential']
        lattice = line_lattice.uniform_parameters(uniform_interaction=t, uniform_onsite_potential=v)
        fhm = FermiHubbardModel(lattice=lattice, onsite_interaction=u)
        self.lmp = LatticeModelProblem(fhm)

    def initial_state(self):
        return {}

    def inputs(self):
        """TODO: Take in instances of bigraphs and convert and then fit them into the Fermi Hubbard Model."""
        return {}

    def outputs(self):
        return {'ground_state': 'float'}

    def update(self, state, interval):
        numpy_solver = NumPyMinimumEigensolver()
        qubit_mapper = JordanWignerMapper()
        calc = GroundStateEigensolver(qubit_mapper, numpy_solver)
        res = calc.solve(self.lmp)
        print(res)
        # TODO: somehow be able to create a new instance from this ground state.
        return {'ground_state': res}



# -- step and process implementation references for registration --
from bsp.data_model.base import Implementation


# TODO: add the rest of the implementations here

COBRA_PROCESSES = [
    Implementation(
        address='cobra-process',
        location='processes.cobra_process.CobraProcess',
        dependencies=["cobra", "imageio"]
    ),
    Implementation(
        address='sed-cobra-process',
        location='processes.cobra_process.SedCobraProcess',
        dependencies=["cobra", "imageio"]
    ),
    Implementation(
        address='dynamic-fba',
        location='processes.cobra_process.DynamicFBA',
        dependencies=["cobra", "imageio"]

    )
]

COPASI_PROCESSES = [
    Implementation(
        address='copasi-process',
        location='processes.copasi_process.CopasiProcess',
        dependencies=["copasi-basico"]
    ),
    Implementation(
        address='sed-copasi-process',
        location='processes.copasi_process.SedCopasiProcess',
        dependencies=["copasi-basico"]
    )
]


MEM3DG_PROCESSES = [
    # Implementation(
    #     address='membrane-process',
    #     location='processes.membrane_process.MembraneProcess',
    #     dependencies=["conda-forge::pymem3dg"]
    # ),
    Implementation(
        address='simple-membrane-process',
        location='processes.simple_membrane_process.SimpleMembraneProcess',
        dependencies=["conda-forge::pymem3dg", "netCDF4"]
    )
]


SMOLDYN_PROCESSES = [
    Implementation(
        address='smoldyn-process',
        location='processes.smoldyn_process.SmoldynProcess',
        dependencies=["smoldyn"]
    )
]


TELLURIUM_PROCESSES = [
    Implementation(
        address='tellurium-process',
        location='processes.tellurium_process.TelluriumProcess',
        dependencies=["tellurium"]
    )
]

PROCESS_IMPLEMENTATIONS = COBRA_PROCESSES + COPASI_PROCESSES + MEM3DG_PROCESSES + SMOLDYN_PROCESSES + TELLURIUM_PROCESSES


STEP_IMPLEMENTATIONS = [
    Implementation(
        address='mongo-emitter',
        location='steps.main_steps.MongoDatabaseEmitter',
        dependencies=["pymongo"]
    )
    # ('output-generator', 'steps.main_steps.OutputGenerator'),
    # ('time-course-output-generator', 'steps.main_steps.TimeCourseOutputGenerator'),
    # ('smoldyn_step', 'steps.main_steps.SmoldynStep', ['smoldyn']),
    # ('simularium_smoldyn_step', 'steps.main_steps.SimulariumSmoldynStep', ['smoldyn', 'simulariumio']),
]

INITIAL_MODULES = PROCESS_IMPLEMENTATIONS



"""
Data model relating to SED2 (S.E.D.) specification
"""


from dataclasses import dataclass
from typing import *
from types import NoneType
from abc import abstractmethod, ABC

import requests
from pydantic import Field, create_model

from bsp.data_model.base import BaseModel, BaseClass as _BaseClass


@dataclass
class SimulationModelParameter(_BaseClass):
    """
        Attributes:
            name:`str`
            feature:`str`
            value:`Union[float, int, str]`
            scope:`str`
    """
    name: str
    value: Union[float, int, str]
    feature: str
    scope: str = None


# TODO: what could this specific name be? Are these changes etc, unique to Time Course / '2d' simulation?
@dataclass
class SpeciesChange(_BaseClass):  # <-- this is done like set_species('B', kwarg=) where the inner most keys are the kwargs
    species_name: str
    unit: Union[str, NoneType, SimulationModelParameter] = None
    initial_concentration: Optional[Union[float, SimulationModelParameter]] = None
    initial_particle_number: Optional[Union[float, NoneType, SimulationModelParameter]] = None
    initial_expression: Union[str, NoneType, SimulationModelParameter] = None
    expression: Union[str, NoneType, SimulationModelParameter] = None


@dataclass
class GlobalParameterChange(_BaseClass):  # <-- this is done with set_parameters(PARAM, kwarg=). where the inner most keys are the kwargs
    parameter_name: str
    initial_value: Union[float, NoneType] = None
    initial_expression: Union[str, NoneType] = None
    expression: Union[str, NoneType] = None
    status: Union[str, NoneType] = None
    param_type: Union[str, NoneType] = None  # ie: fixed, assignment, reactions, etc


@dataclass
class ReactionParameter(_BaseClass):
    parameter_name: str
    value: Union[float, int, str]


@dataclass
class ReactionChange(_BaseClass):
    """
        Attributes:
            reaction_name:`str`: name of the reaction you wish to change.
            parameter_changes:`List[ReactionParameter]`: list of parameters you want to change from
                `reaction_name`. Defaults to `[]`, which denotes no parameter changes.

    """
    reaction_name: str
    parameter_changes: List[ReactionParameter] = None
    reaction_scheme: Union[NoneType, str] = None


@dataclass
class TimeCourseModelChanges(_BaseClass):
    species_changes: List[SpeciesChange] = None
    global_parameter_changes: List[GlobalParameterChange] = None
    reaction_changes: List[ReactionChange] = None


@dataclass
class ModelSource(_BaseClass):
    value: str

    @abstractmethod
    def validate_source(self):
        pass


@dataclass
class BiomodelID(ModelSource):
    def __init__(self, value):
        super().__init__(value)
        self.validate_source()

    def validate_source(self):
        if 'BIO' not in self.value:
            raise AttributeError('You must pass a valid biomodel id.')


@dataclass
class ModelFilepath(ModelSource):
    def __init__(self, value):
        super().__init__(value)
        self.validate_source()

    def validate_source(self):
        if '/' not in self.value:
            raise AttributeError('You must pass a valid model path.')


@dataclass
class ModelChange(_BaseClass):
    name: str
    scope: str
    value: Dict


@dataclass
class ModelChanges(_BaseClass):
    species_changes: List[ModelChange] = None
    param_changes: List[ModelChange] = None
    reaction_changes: List[ModelChange] = None


class ModelUnits:
    def __init__(self, **units_config):
        for k, v in units_config:
            self.__setattr__(k, v)


@dataclass
class SedModel(_BaseClass):
    """The data model declaration for process configuration schemas that support SED.

        Attributes:
            model_id: `str`
            model_source: `Union[biosimulators_processes.data_model.ModelFilepath, biosimulators_processes.data_model.BiomodelId]`
            model_language: `str`
            model_name: `str`
            model_changes: `biosimulators_processes.data_model.TimeCourseModelChanges`
    """
    model_source: Union[BiomodelID, ModelFilepath, str]
    model_id: str = None
    model_name: str = None
    model_language: str = Field(default='sbml')
    model_changes: ModelChanges = None
    model_units: ModelUnits = None

    def set_id(self, model_id=None):
        if model_id is None:
            if isinstance(self.model_source, ModelFilepath) \
                    or isinstance(self.model_source, str) and '/' in self.model_source:
                if isinstance(self.model_source, ModelFilepath):
                    source = self.model_source.value
                else:
                    source = self.model_source
                modId = source.split('/')[-1]
            elif isinstance(self.model_source, str):
                modId = self.model_source
            else:
                modId = self.model_source.value
            return f'model_from_{modId}'
        else:
            return model_id

    def set_name(self, name=None):
        if name is None:
            if self.model_id is not None:
                return self.model_id
            else:
                return 'Un-named model'
        else:
            return name

    def set_source(self):
        if isinstance(self.model_source, str):
            if 'BIOMD' in self.model_source.upper():
                self.model_source = BiomodelID(value=self.model_source)
            elif '/' in self.model_source:
                self.model_source = ModelFilepath(value=self.model_source)
            else:
                raise AttributeError('You must pass either a valid model filepath or valid Biomodel id.')

    def get_model_source_info(self, source: str, validator, **kwargs):
        """Currently only support BioModel id."""
        return validator(source, **kwargs)

    def get_biomodel_model_source_info(self, biomodel_id: str) -> Dict:
        """Return information about the BioModel ID passed."""
        return requests.get(
            url=f'https://www.ebi.ac.uk/biomodels/{biomodel_id}',
            headers={'accept': 'application/json'}).json()

    def set_model_source_info(self, **kwargs):
        """Currently only support biomodel id"""
        return self.get_biomodel_model_source_info(**kwargs)


# TODO: Provide this model if 'CopasiProcess', etc is selected by the user in prompt.

@dataclass
class TimeCourseModel(SedModel):
    """The data model declaration for process configuration schemas that support SED.

        Attributes:
            model_id: `str`
            model_source: `Union[biosimulators_processes.data_model.ModelFilepath, biosimulators_processes.data_model.BiomodelId]`
            model_language: `str` defaults to sbml
            model_name: `str`
            model_changes: `biosimulators_processes.data_model.TimeCourseModelChanges`
    """
    model_language: str = 'sbml',
    model_changes: TimeCourseModelChanges = None,
    model_units: ModelUnits = None

    def __init__(self,
                 model_source: Union[BiomodelID, ModelFilepath, str],
                 model_changes=model_changes,
                 model_id=None,
                 model_name=None):
        """Class which inherits SedModel."""
        super().__init__(model_source, model_id, model_name, model_changes)
        # TODO: extract functionality for algorithms related to UTC sims
        self.model_id = self.set_id(model_id)
        self.model_name = self.set_name(model_name)
        source_id = self.model_source.value \
            if isinstance(self.model_source, BiomodelID) or isinstance(self.model_source, ModelFilepath) \
            else self.model_source
        if 'BIOMD' in source_id:
            self.source_info = self.get_biomodel_model_source_info(source_id)


class SteadyStateModel(SedModel):
    def __init__(self,
                 model_source: Union[BiomodelID, ModelFilepath, str],
                 model_id=None,
                 model_name=None,
                 model_language: str = None,
                 model_changes: ModelChanges = None,
                 model_units: ModelUnits = None):
        """Class which inherits SedModel. # TODO: expand this."""
        super().__init__(model_source, model_id, model_name, model_language, model_changes, model_units)
        self.model_id = self.set_id(model_id)
        self.model_name = self.set_name(model_name)


class SpatialModel(SedModel):
    def __init__(self,
                 model_source: Union[BiomodelID, ModelFilepath, str],
                 model_id=None,
                 model_name=None,
                 model_language: str = None,
                 model_changes: ModelChanges = None,
                 model_units: ModelUnits = None):
        """Class which inherits SedModel. # TODO: expand this."""
        super().__init__(model_source, model_id, model_name, model_language, model_changes, model_units)
        self.model_id = self.set_id(model_id)
        self.model_name = self.set_name(model_name)


@dataclass
class TimeCourseProcess(_BaseClass):
    """Used as config for BioBuilder API"""
    model: SedModel
    method: str = 'lsoda'
    model_language: str = 'sbml'


@dataclass
class ExperimentMetadata(_BaseClass):
    def __init__(self):
        raise NotImplementedError('Please do not use this class yet. Thanks!')


@dataclass
class SimulationResult(_BaseClass):
    value: Dict[str, Any]

    def serialize(self):
        pass


@dataclass
class Experiment:
    """Consider this a 'simulation run'."""
    simulation_model: Union[TimeCourseModel, Dict[str, Any]]  # where the theory/hypothesis lies. you're stating your biological case here.
    duration: Union[int, float]  # one level above theory model. you're sure of your case and want to not tell but show it.
    results: Dict[str, Union[NoneType, SimulationResult]] = None  # consider async
    name: str = None
    metadata: ExperimentMetadata = None

    def get_results(self, source: Any):
        return source.get('results', {})

    def set_results(self):
        source = None
        results = self.get_results(source)
        self.results = results


def dynamic_process_config(name: str = None, config: Dict = None, **kwargs):
    config = config or {}
    config.update(kwargs)
    dynamic_config_types = {}
    for param_name, param_val in config.items():
        dynamic_config_types[param_name] = (type(param_val), ...)

    model_name = 'ProcessConfig'
    if name is not None:
        proc_name = name.replace(name[0], name[0].upper())
        dynamic_name = proc_name + model_name
    else:
        dynamic_name = model_name

    DynamicProcessConfig = create_model(__model_name=dynamic_name, **dynamic_config_types)

    return DynamicProcessConfig(**config)


class Port(_BaseClass):
    value: Dict






import numpy as np
from typing import *

from simulariumio import TrajectoryData, JsonWriter, BinaryWriter
from simulariumio.filters import TranslateFilter
from simulariumio.smoldyn import SmoldynData, SmoldynConverter


# --agent display --

def calculate_agent_radius(m: float, rho: float, scaling_factor: float = 10**(-2)) -> float:
    """Calculate the radius of an agent given its molecular mass and density. Please note: the molecular mass
        of MinE is 11000 Da with a protein density of 1.35 g/cm^3 (1350 kg/m^3).

        Args:
            m:`float`: the molecular mass of the given agent/particle (Daltons).
            rho:`float`: the density of the given agent/particle (kg/m^3).
            scaling_factor:`float`: tiny number by which to scale the output measurement. Defaults to
                `10**(-2)`, which effectively converts from nm to cm.

        Returns:
            `float`: radius of the given agent.
    """
    dalton_to_kg = 1.66053906660e-27  # Conversion factor from Daltons to kilograms
    m_kg = m * dalton_to_kg  # Convert mass to kilograms
    radius_m = ((3 * m_kg) / (4 * np.pi * rho)) ** (1 / 3)  # Calculate radius in meters
    radius_nm = radius_m * 1e9  # Convert radius to nanometers
    return radius_nm * scaling_factor


def generate_agent_params(
        species_names: List[str],
        global_density: Optional[float] = None,
        basis_m: Optional[int] = None,
        model_fp: Optional[str] = None,
        **config
) -> Dict:
    """Generate a dictionary of agent parameters for the purpose of simulation input configuration which define the
        molecular mass and density inherent to a given agent based species names. We cannot call the species
        names from the smoldyn model file here directly, so you MUST pass them in here.

        Args:
            species_names:`List[str]`: List of species names which coorespond to the species names
                in the relative smoldyn model file being simulated.
            global_density:`Optional[float]`: Density by which all agent densities are set. NOTE: this value is
              required if not passing explicit agent configs (**config). Defaults to `None`.
            basis_m:`Optional[int]`: Upper bound value of the molecular mass by which to set the basis for the
              randomization of agent molecular masses in the `randomize_mass` function which takes in a basis
              integer and returns a random integer between 0 and that number. NOTE: this value is required
              if not passing explicit agent configs (**config). Defaults to `None`.
            model_fp:`Optional[str]`: path to a smoldyn model file. Defaults to `None`.
        Keyword Args:
            <AGENT_NAME>:`dict`: an agent name (which should match that which is returned by
                smoldyn.Simulation.getSpeciesName()) as the definition (for example: `MinE=`) and a dictionary with
                `'molecular_mass'` and `'density'` as the keys.

    """
    params = {}
    if not config and not global_density and not basis_m:
        raise ValueError(
            f'You must pass either keyword arguments where the keyword is the agent name and the value is a dict defining molecular_mass and density OR a density AND basis molecular mass.'
        )
    if not config:
        for name in species_names:
            agent_config = config.get(f'{name}')
            if agent_config:
                mass = agent_config['molecular_mass']
                density = agent_config['density']
            else:
                mass = randomize_mass(basis_m)
                density = global_density
            params[name] = {
                'density': density,
                'molecular_mass': mass
            }
    else:
        for k in config.keys():
            params[k] = config[k]
    return params


def randomize_mass(origin: float) -> int:
    return np.random.randint(int(origin))


# -- data composition --

def translate_data_object(
    data: SmoldynData,
    box_size: float,
    n_dim=3,
    translation_magnitude: Optional[Union[int, float]] = None
) -> TrajectoryData:
    """Translate the data object's data if the coordinates are all positive to center the data in the
           simularium viewer.

           Args:
               data:`SmoldynData`: configured simulation output data instance.
               box_size: size of the simularium viewer box.
               n_dim: n dimensions of the simulation output. Defaults to `3`.
               translation_magnitude: magnitude by which to translate and filter. Defaults to `-box_size / 2`.

           Returns:
               `TrajectoryData`: translated data object instance.
       """
    translation_magnitude = translation_magnitude or -box_size / 2
    return SmoldynConverter(data).filter_data(
        [TranslateFilter(translation_per_type={}, default_translation=translation_magnitude * np.ones(n_dim))]
    )


# -- io --

def write_simularium_file(
    data: Union[SmoldynData, TrajectoryData],
    simularium_fp: str,
    json: bool = True,
    validate: bool = True
) -> None:
    """Takes in either a `SmoldynData` or `TrajectoryData` instance and saves a simularium file based on it
        with the name of `simularium_filename`.

        Args:
            data(:obj:`Union[SmoldynData, TrajectoryData]`): data object to save.
            simularium_fp(:obj:`str`): name by which to save the new simularium file.
            json(:obj:`bool`): exports simularium file in JSON format if true; exports in binary if false. Defaults
                to `False` for optimization's sake.
            validate(:obj:`bool`): whether to call the wrapped method using `validate_ids=True`. Defaults
                to `True`.
    """
    print(f'Writing simularium file "{simularium_fp}"')
    if json:
        writer = JsonWriter()
    else:
        writer = BinaryWriter()
    return writer.save(trajectory_data=data, output_path=simularium_fp, validate_ids=validate)


def read_smoldyn_simulation_configuration(filename):
    ''' Read a configuration for a Smoldyn simulation

    Args:
        filename (:obj:`str`): path to model file

    Returns:
        :obj:`list` of :obj:`str`: simulation configuration
    '''
    with open(filename, 'r') as file:
        return [line.strip('\n') for line in file]


def write_smoldyn_simulation_configuration(configuration, filename):
    ''' Write a configuration for Smoldyn simulation to a file

    Args:
        configuration
        filename (:obj:`str`): path to save configuration
    '''
    with open(filename, 'w') as file:
        for line in configuration:
            file.write(line)
            file.write('\n')


def add_output_commands(model_fp, duration):
    config = read_smoldyn_simulation_configuration(model_fp)
    has_output_commands = any([v.startswith('output') for v in config])
    if not has_output_commands:
        cmd_i = 0
        for i, v in enumerate(config):
            if v == 'end_file':
                cmd_i += i - 1
            stop_key = 'time_stop'

            if f'define {stop_key.upper()}' in v:
                new_v = f'define TIME_STOP   {duration}'
                config.remove(config[i])
                config.insert(i, new_v)
            elif v.startswith(stop_key):
                new_v = f'time_stop TIME_STOP'
                config.remove(config[i])
                config.insert(i, new_v)
        cmds = ["output_files modelout.txt",
                "cmd i 0 TIME_STOP 2 executiontime modelout.txt",
                "cmd i 0 TIME_STOP 2 listmols modelout.txt"]
        current = cmd_i
        for cmd in cmds:
            config.insert(current, cmd)
            current += 1

        write_smoldyn_simulation_configuration(config, model_fp)

    out_file = model_fp.replace(model_fp.split('/')[-1].split('.')[0], 'modelout')
    return out_file

# 1. in worker, read in path params from job_params
# 2. download file to get local temp filepath
# 3. get ouput filepath as add_output_commands() (rewrite the files if need be locally)
# 4. reupload/replace the bucket file with the newly created local file
# 5. Continue to run process



"""
Type schema definitions relating to process implementations within this application.

The type name suffix is used to denote its function. Currently there are two suffixes:

- ...Type -> this relates to the schema for input and output port data only
"""


__all__ = [
    'BoundsType',
    'PositiveFloatType',
    'MechanicalForcesType',
    'ProteinDensityType',
    'GeometryType',
    'VelocitiesType',
    'OsmoticParametersType',
    'SurfaceTensionParametersType',
    'ParticleType'
]


# create new types
def apply_non_negative(schema, current, update, core):
    new_value = current + update
    return max(0, new_value)


# -- types: that is, schemas related to input and output port data, not configs

PositiveFloatType = {
    '_type': 'positive_float',
    '_inherit': 'float',
    '_apply': apply_non_negative
}

BoundsType = {
    'lower': 'maybe[float]',
    'upper': 'maybe[float]'
}

ForcesType = 'list[list[float]]'

MechanicalForcesType = {
    '_inherit': ForcesType
}

GeometryType = {
    "faces": 'list',  # "list[list[integer]]",
    "vertices": 'list'  # "list[list[float]]",
}

OsmoticParametersType = {
    'preferred_volume': 'float',
    'volume': 'maybe[float]',  # current (actual) volume by which a potential is created when compared to preferred volume
    'strength': 'float',
    'reservoir_volume': 'float'  # represents the environment outside the membrane, which can exchange solutes or exert osmotic pressure on the system
}

SurfaceTensionParametersType = {
    'area': 'float',  # use Geometry.getArea() for this value
    'modulus': 'float',  # static value (in constructor?)
    'preferred_area': 'float'
}

ProteinDensityType = 'list[float]'

VelocitiesType = 'list[float]'

ParticleType = {
    'coordinates': 'list',
    'species_id': 'string',
    'state': 'integer'
}

